/*******************************************************************************
*                                 ZM4xx
*                       ----------------------------
*                       innovating embedded platform
*
* Copyright (c) 2001-present Guangzhou ZHIYUAN Electronics Co., Ltd.
* ALL rights reserved.
*
* THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
* IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
* FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
*
* The software is free to use in a commercial product and should not be
* redistributed in any way. See the LICENSE file for more details.
*
* Contact information:
* web site:    http://www.zlg.cn/
*******************************************************************************/

/**
 * \file
 * \brief  SX126x FSK driver
 *
 * \internal
 * \par Modification history
 * - v1.2.5 22-12-27
 * - v1.2.4 22-11-17
 * - v1.2.3 22-08-24
 * - v1.2.2 22-08-22
 * - v1.1.6 22-05-16
 * - v1.1.3 21-09-15
 * - v1.1.1 21-08-03
 * - v1.1.0 21-01-26
 * - v1.0.0 20-09-08  sup, first implementation.
 * \endinternal
 */


#include <math.h>
#include "string.h"
#include "stdio.h"
#include "sx126x_fsk.h"


#define SM_STATE_SET(p_this, state)   ((p_this)->run_state.sm_state = state)
#define SM_STATE_GET(p_this)          ((p_this)->run_state.sm_state)


static int  __sx126x_sync_id_set (void *p_arg, uint8_t *p_id, uint8_t len);
static int  __sx126x_mode_set (void *p_arg, uint8_t mode);
static int  __sx126x_cur_rssi_get (void *p_arg, int16_t *p_rssi);
static int  __sx126x_freq_set (void *p_arg, uint32_t frequency);
static int  __sx126x_set_rf_power (radio_sx126x_fsk_dev_t *p_this,
                                   int8_t                  power,
                                   uint8_t                 ramp_time);
static void __sx126x_set_rx (radio_sx126x_fsk_dev_t *p_this, uint32_t timeout);
static void __sx126x_set_dio_irq_params (radio_sx126x_fsk_dev_t *p_this,
                                         uint16_t                irq_mask,
                                         uint16_t                dio1_mask,
                                         uint16_t                dio2_mask,
                                         uint16_t                dio3_mask);
static int __sx126x_set_standby (radio_sx126x_fsk_dev_t *p_this, uint8_t standby_mode);
static void __sx126x_set_fsk_mode (void *p_drv);


static uint8_t __g_syncword_len = 0;
static const sx126x_fsk_bw_t __g_sx126x_bandwidth[] = {
    {4800,   FSK_RX_BW_4800},
    {5800,   FSK_RX_BW_5800},
    {7300,   FSK_RX_BW_7300},
    {9700,   FSK_RX_BW_9700},
    {11700,  FSK_RX_BW_11700},
    {14600,  FSK_RX_BW_14600},
    {19500,  FSK_RX_BW_19500},
    {23400,  FSK_RX_BW_23400},
    {29300,  FSK_RX_BW_29300},
    {39000,  FSK_RX_BW_39000},
    {46900,  FSK_RX_BW_46900},
    {58600,  FSK_RX_BW_58600},
    {78200,  FSK_RX_BW_78200},
    {93800,  FSK_RX_BW_93800},
    {117300, FSK_RX_BW_117300},
    {156200, FSK_RX_BW_156200},
    {187200, FSK_RX_BW_187200},
    {234300, FSK_RX_BW_234300},
    {312000, FSK_RX_BW_312000},
    {373600, FSK_RX_BW_373600},
    {467000, FSK_RX_BW_467000},
};

static const sx126x_fsk_rate_level_t __g_sx126x_rate_lv[] = {
    {0,      0,      0},  /* Ч */
    {30000,  20000,  46900},
    {60000,  20000,  93800},
    {90000,  30000,  93800},
    {120000, 40000,  156200},
    {150000, 50000,  156200},
    {180000, 60000,  187200},
    {210000, 70000,  234300},
    {240000, 80000,  312000},
    {270000, 100000, 373600},
    {300000, 100000, 373600},
};


static void __sx126x_delay_ms (radio_sx126x_fsk_dev_t *p_this, uint32_t ms)
{
    p_this->p_delay->pfn_delay_ms(ms);
}

#if 0
static void __sx126x_delay_us (radio_sx126x_fsk_dev_t *p_this, uint32_t us)
{
    p_this->p_delay->pfn_delay_us(us);
}
#endif

static void __send_byte (radio_sx126x_fsk_dev_t *p_this, uint8_t byte)
{
    p_this->p_spi->pfn_spi_write_byte(byte);
}

static uint8_t __recv_byte (radio_sx126x_fsk_dev_t *p_this)
{
    return p_this->p_spi->pfn_spi_read_byte();
}


static void __sx126x_wait_on_busy (radio_sx126x_fsk_dev_t *p_this)
{
    while(1) {
        if (p_this->p_gpio->pfn_busy_pin_read() == 0) {
            break;
        }
    }
}


static void __sx126x_wake_up (radio_sx126x_fsk_dev_t *p_this)
{
    uint32_t key;
    uint8_t  status   = 0;
    uint8_t  cpu_busy = 0;

    key = p_this->p_lock->pfn_cpu_lock();

    do {
        p_this->p_gpio->pfn_sel_pin_set(0);
        __send_byte(p_this, RADIO_GET_STATUS);
        __send_byte(p_this, 0x00);
        status = __recv_byte(p_this);
        cpu_busy = (status & 0x80) >> 7;
        p_this->p_gpio->pfn_sel_pin_set(1);
    } while (cpu_busy == 1);

    p_this->p_lock->pfn_cpu_unlock(key);
    __sx126x_wait_on_busy(p_this);
}


static void __sx126x_check_device_ready (radio_sx126x_fsk_dev_t *p_this)
{
    if ((SM_STATE_GET(p_this) == RFLR_STATE_SLEEP) ||
        (SM_STATE_GET(p_this) == RFLR_STATE_RX_DC)) {
        __sx126x_wake_up(p_this);
    }
    __sx126x_wait_on_busy(p_this);
}


static void __sx126x_write_command (radio_sx126x_fsk_dev_t *p_this,
                                    radio_commands_t        command,
                                    uint8_t                *p_buffer,
                                    uint16_t                size)
{
    uint16_t i;
    uint32_t key;

    __sx126x_check_device_ready(p_this);

    key = p_this->p_lock->pfn_cpu_lock();
    p_this->p_gpio->pfn_sel_pin_set(0);
    __send_byte(p_this, (uint8_t)command);

    for (i = 0; i < size; i++) {
        __send_byte(p_this, p_buffer[i]);
    }

    p_this->p_gpio->pfn_sel_pin_set(1);
    p_this->p_lock->pfn_cpu_unlock(key);

    if (command != RADIO_SET_SLEEP) {
        __sx126x_wait_on_busy(p_this);
    }
}

static void __sx126x_read_command (radio_sx126x_fsk_dev_t *p_this,
                                   radio_commands_t        command,
                                   uint8_t                *p_buffer,
                                   uint16_t                size)
{
    uint16_t i;
    uint32_t key;

    __sx126x_check_device_ready(p_this);

    key = p_this->p_lock->pfn_cpu_lock();
    p_this->p_gpio->pfn_sel_pin_set(0);

    __send_byte(p_this, (uint8_t)command);
    __send_byte(p_this, 0x00);

    for (i = 0; i < size; i++) {
        p_buffer[i] = __recv_byte(p_this);
    }

    p_this->p_gpio->pfn_sel_pin_set(1);
    p_this->p_lock->pfn_cpu_unlock(key);

    __sx126x_wait_on_busy(p_this);
}

static void __sx126x_write_registers (radio_sx126x_fsk_dev_t *p_this,
                                      uint16_t                address,
                                      uint8_t                *p_buffer,
                                      uint16_t                size)
{
    uint16_t i;
    uint32_t key;

    __sx126x_check_device_ready(p_this);

    key = p_this->p_lock->pfn_cpu_lock();
    p_this->p_gpio->pfn_sel_pin_set(0);

    __send_byte(p_this, RADIO_WRITE_REGISTER);
    __send_byte(p_this, (address & 0xFF00) >> 8);
    __send_byte(p_this, address & 0x00FF);

    for (i = 0; i < size; i++) {
        __send_byte(p_this, p_buffer[i]);
    }

    p_this->p_gpio->pfn_sel_pin_set(1);
    p_this->p_lock->pfn_cpu_unlock(key);

    __sx126x_wait_on_busy(p_this);
}

static void __sx126x_write_register (radio_sx126x_fsk_dev_t *p_this,
                                     uint16_t                address,
                                     uint8_t                 value)
{
    __sx126x_write_registers(p_this, address, &value, 1);
}

static void __sx126x_read_registers (radio_sx126x_fsk_dev_t *p_this,
                                     uint16_t                address,
                                     uint8_t                *p_buffer,
                                     uint16_t                size)
{
    uint16_t i;
    uint32_t key;

    __sx126x_check_device_ready(p_this);

    key = p_this->p_lock->pfn_cpu_lock();
    p_this->p_gpio->pfn_sel_pin_set(0);

    __send_byte(p_this, RADIO_READ_REGISTER);
    __send_byte(p_this, (address & 0xFF00) >> 8);
    __send_byte(p_this, address & 0x00FF);
    __send_byte(p_this, 0x00);

    for (i = 0; i < size; i++) {
        p_buffer[i] = __recv_byte(p_this);
    }
    p_this->p_gpio->pfn_sel_pin_set(1);
    p_this->p_lock->pfn_cpu_unlock(key);

    __sx126x_wait_on_busy(p_this);
}

static uint8_t __sx126x_read_register (radio_sx126x_fsk_dev_t *p_this,
                                       uint16_t                address)
{
    uint8_t data;
    __sx126x_read_registers(p_this, address, &data, 1);
    return data;
}

static void __sx126x_write_buffer (radio_sx126x_fsk_dev_t *p_this,
                                   uint8_t                 offset,
                                   uint8_t                *p_buffer,
                                   uint8_t                 size)
{
    uint16_t i;
    uint32_t key;

    __sx126x_check_device_ready(p_this);

    key = p_this->p_lock->pfn_cpu_lock();
    p_this->p_gpio->pfn_sel_pin_set(0);

    __send_byte(p_this, RADIO_WRITE_BUFFER);
    __send_byte(p_this, offset);

    for (i = 0; i < size; i++) {
      __send_byte(p_this, p_buffer[i]);
    }

    p_this->p_gpio->pfn_sel_pin_set(1);
    p_this->p_lock->pfn_cpu_unlock(key);

    __sx126x_wait_on_busy(p_this);
}

static void __sx126x_read_buffer (radio_sx126x_fsk_dev_t *p_this,
                                  uint8_t                 offset,
                                  uint8_t                *p_buffer,
                                  uint8_t                 size)
{
    uint16_t i;
    uint32_t key;

    __sx126x_check_device_ready(p_this);

    key = p_this->p_lock->pfn_cpu_lock();
    p_this->p_gpio->pfn_sel_pin_set(0);

    __send_byte(p_this, RADIO_READ_BUFFER);
    __send_byte(p_this, offset);
    __send_byte(p_this, 0x00);

    for (i = 0; i < size; i++) {
        p_buffer[i] = __recv_byte(p_this);
    }

    p_this->p_gpio->pfn_sel_pin_set(1);
    p_this->p_lock->pfn_cpu_unlock(key);

    __sx126x_wait_on_busy(p_this);
}


static uint16_t __crc16_ccitt (uint8_t *p_buf, uint32_t len)
{
    uint16_t ccitt16 = 0x1021;
    uint16_t crc     = 0;
    int      i, j;
    for (i = 0; i < len; i++) {
        crc ^= (p_buf[i] << 8);
        for (j = 0; j < 8; j++) {
            if (crc & 0x8000) {
                crc <<= 1;
                crc ^= ccitt16;
            } else {
                crc <<= 1;
            }
        }
    }
    return crc;
}



static int __sx126x_set_cap_val_of_xtal (radio_sx126x_fsk_dev_t *p_this)
{
    uint8_t cap_val = p_this->p_devinfo->p_fsk_param->dev_type.cap_val;
    if (cap_val > 47) {
        return RADIO_RET_CFG_OUT_OF_RANGE;
    }
    __sx126x_set_standby(p_this, STDBY_XOSC_MODE);
    __sx126x_write_register(p_this, REG_XTA_TRIM, cap_val);
    __sx126x_write_register(p_this, REG_XTB_TRIM, cap_val);
    return RADIO_RET_OK;
}


static int __sx126x_set_fs (radio_sx126x_fsk_dev_t *p_this)
{
    __sx126x_write_command(p_this, RADIO_SET_FS, 0, 0);
    SM_STATE_SET(p_this, RFLR_STATE_IDLE);
    return RADIO_RET_OK;
}



static int __sx126x_set_rx_duty_cycle (radio_sx126x_fsk_dev_t *p_this,
                                       uint32_t                rx_time,
                                       uint32_t                sleep_time)
{
    uint8_t buf[6] = {0};

    buf[0] = (uint8_t)((rx_time >> 16) & 0xFF);
    buf[1] = (uint8_t)((rx_time >> 8) & 0xFF);
    buf[2] = (uint8_t)(rx_time & 0xFF);
    buf[3] = (uint8_t)((sleep_time >> 16) & 0xFF);
    buf[4] = (uint8_t)((sleep_time >> 8) & 0xFF);
    buf[5] = (uint8_t)(sleep_time & 0xFF);
    __sx126x_write_command(p_this, RADIO_SET_RXDUTYCYCLE, buf, 6);
    SM_STATE_SET(p_this, RFLR_STATE_RX_RUNNING);

    return RADIO_RET_OK;
}

#if 0
static uint32_t __sx126x_get_random (radio_sx126x_fsk_dev_t *p_this)
{
    uint8_t buf[4] = {0};

    __sx126x_set_rx(p_this, 0xFFFFFF);
    __sx126x_delay_ms(p_this, 1);
    __sx126x_read_registers(p_this, REG_RAND_NUM_GEN, buf, 4);
    __sx126x_set_standby(p_this, STDBY_XOSC_MODE);

    return (buf[0] << 24) | (buf[1] << 16) | (buf[2] << 8) | buf[3];
}



static void __sx126x_set_tx_infinite_preamble (radio_sx126x_fsk_dev_t *p_this)
{
    __sx126x_write_command(p_this, RADIO_SET_TXCONTINUOUSPREAMBLE, 0, 0);
}


static void __sx126x_set_rx_tx_fallback_mode (radio_sx126x_fsk_dev_t *p_this,
                                              uint8_t                 fallback_mode)
{
    __sx126x_write_command(p_this, RADIO_SET_TXFALLBACKMODE, &fallback_mode, 1);
}



static uint8_t __sx126x_get_status (radio_sx126x_fsk_dev_t *p_this)
{
    uint8_t status  = 0;
    __sx126x_read_command(p_this, RADIO_GET_STATUS, (uint8_t *)&status, 1);
    return status;
}


static uint16_t __sx126x_get_device_errors (radio_sx126x_fsk_dev_t *p_this)
{
    uint16_t error_val;
    __sx126x_read_command(p_this, RADIO_GET_ERROR, (uint8_t *)&error_val, 2);
    return error_val;
}


static uint32_t __sx126x_radio_random (void *p_drv)
{
    radio_sx126x_fsk_dev_t *p_this = (radio_sx126x_fsk_dev_t *)p_drv;
    uint8_t                 i      = 0;
    uint32_t                rnd    = 0;
    int16_t                 rssi   = 0;

    __sx126x_set_fsk_mode(p_this);
    __sx126x_set_rx(p_this, 0xFFFFFF);

    for (i = 0; i < 32; i++) {
        __sx126x_delay_ms(p_this, 1);
        __sx126x_cur_rssi_get(p_this, &rssi);
        rnd |= ((uint32_t)rssi & 0x01) << i;
    }

    __sx126x_mode_set(p_this, SLEEP_MODE);
    return rnd;
}


static void __sx126x_set_rx (radio_sx126x_fsk_dev_t *p_this, uint32_t timeout)
{
    uint8_t buf[3] = {0};

    buf[0] = (uint8_t)((timeout >> 16) & 0xFF);
    buf[1] = (uint8_t)((timeout >> 8) & 0xFF);
    buf[2] = (uint8_t)(timeout & 0xFF);
    __sx126x_write_command(p_this, RADIO_SET_RX, buf, 3);
    SM_STATE_SET(p_this, RFLR_STATE_RX_RUNNING);
}



#endif

static void __sx126x_clear_device_errors (radio_sx126x_fsk_dev_t *p_this)
{
    uint8_t buf[2] = {0};
    __sx126x_write_command(p_this, RADIO_CLR_ERROR, buf, 2);
}



static void __sx126x_set_rx_boosted (radio_sx126x_fsk_dev_t *p_this, uint32_t timeout)
{
    uint8_t buf[3] = {0};

    __sx126x_write_register(p_this, REG_RX_GAIN, 0x96);

    buf[0] = (uint8_t)((timeout >> 16) & 0xFF);
    buf[1] = (uint8_t)((timeout >> 8) & 0xFF);
    buf[2] = (uint8_t)(timeout & 0xFF);
    __sx126x_write_command(p_this, RADIO_SET_RX, buf, 3);
    SM_STATE_SET(p_this, RFLR_STATE_RX_RUNNING);
}


static int __sx126x_radio_rx_boosted (void *p_drv, uint32_t timeout)
{
    radio_sx126x_fsk_dev_t *p_this = (radio_sx126x_fsk_dev_t *)p_drv;

    __sx126x_set_dio_irq_params(p_this,
                                IRQ_RADIO_ALL,
                                IRQ_RADIO_ALL,
                                IRQ_RADIO_NONE,
                                IRQ_RADIO_NONE);

    if (p_this->p_devinfo->p_fsk_param->rx_mode == 0) {
        __sx126x_set_rx_boosted(p_this, 0xFFFFFF);
    } else {
        __sx126x_set_rx_boosted(p_this, timeout << 6);
    }

    return RADIO_RET_OK;
}

static void __sx126x_set_stop_rxtimer_on_preamble_detect (radio_sx126x_fsk_dev_t *p_this,
                                                          bool                    enable)
{
    __sx126x_write_command(p_this, RADIO_SET_STOPRXTIMERONPREAMBLE, (uint8_t*)&enable, 1);
}

static void __sx126x_calibrate (radio_sx126x_fsk_dev_t *p_this, uint8_t calib_param)
{
    __sx126x_write_command(p_this, RADIO_CALIBRATE, (uint8_t *)&calib_param, 1);
}

static void __sx126x_set_regulator_mode (radio_sx126x_fsk_dev_t *p_this, uint8_t mode)
{
    __sx126x_write_command(p_this, RADIO_SET_REGULATORMODE, (uint8_t *)&mode, 1);
}

static void __sx126x_set_packet_type (radio_sx126x_fsk_dev_t *p_this, uint8_t packet_type)
{
    __sx126x_write_command(p_this, RADIO_SET_PACKETTYPE, (uint8_t *)&packet_type, 1);
}

static uint8_t __sx126x_get_packet_type (radio_sx126x_fsk_dev_t *p_this)
{
    uint8_t packet_type = 0;
    __sx126x_read_command(p_this, RADIO_GET_PACKETTYPE, (uint8_t *)&packet_type, 1);
    return packet_type;
}

static void __sx126x_calibrate_image (radio_sx126x_fsk_dev_t *p_this, uint32_t freq)
{
    uint8_t cal_freq[2] = {0};

    if (freq > 900000000) {
        cal_freq[0] = 0xE1;
        cal_freq[1] = 0xE9;
    } else if (freq > 850000000) {
        cal_freq[0] = 0xD7;
        cal_freq[1] = 0xDB;
    } else if (freq > 770000000) {
        cal_freq[0] = 0xC1;
        cal_freq[1] = 0xC5;
    } else if (freq > 460000000) {
        cal_freq[0] = 0x75;
        cal_freq[1] = 0x81;
    } else if (freq > 425000000) {
        cal_freq[0] = 0x6B;
        cal_freq[1] = 0x6F;
    }

    __sx126x_write_command(p_this, RADIO_CALIBRATEIMAGE, cal_freq, 2);
}

static void __sx126x_set_dio2_as_rf_switch_ctrl (radio_sx126x_fsk_dev_t *p_this,
                                                 uint8_t                 enable)
{
    __sx126x_write_command(p_this, RADIO_SET_RFSWITCHMODE, &enable, 1);
}

static void __sx126x_set_dio3_as_tcxo_ctrl (radio_sx126x_fsk_dev_t *p_this,
                                            uint8_t                 tcxo_voltage,
                                            uint32_t                timeout)
{
    uint8_t buf[4] = {0};

    buf[0] = tcxo_voltage & 0x07;
    buf[1] = (uint8_t)((timeout >> 16) & 0xFF);
    buf[2] = (uint8_t)((timeout >> 8) & 0xFF);
    buf[3] = (uint8_t)(timeout & 0xFF);

    __sx126x_write_command(p_this, RADIO_SET_TCXOMODE, buf, 4);
}

static void __sx126x_set_buffer_base_address (radio_sx126x_fsk_dev_t *p_this,
                                              uint8_t                 tx_base_address,
                                              uint8_t                 rx_base_address)
{
    uint8_t buf[2] = {0};

    buf[0] = tx_base_address;
    buf[1] = rx_base_address;
    __sx126x_write_command(p_this, RADIO_SET_BUFFERBASEADDRESS, buf, 2);
}

static void __sx126x_set_fsk_mode (void *p_drv)
{
    radio_sx126x_fsk_dev_t *p_this = (radio_sx126x_fsk_dev_t *)p_drv;
    __sx126x_set_packet_type(p_this, PACKET_TYPE_GFSK);
}

static int __sx126x_set_standby (radio_sx126x_fsk_dev_t *p_this, uint8_t standby_mode)
{
    __sx126x_write_command(p_this, RADIO_SET_STANDBY, (uint8_t*)&standby_mode, 1);
    SM_STATE_SET(p_this, RFLR_STATE_IDLE);
    return RADIO_RET_OK;
}


static void __sx126x_set_pa_config (radio_sx126x_fsk_dev_t *p_this,
                                    uint8_t                 pa_duty_cycle,
                                    uint8_t                 hp_max,
                                    uint8_t                 device_sel,
                                    uint8_t                 pa_lut)
{
    uint8_t buf[4] = {0};

    buf[0] = pa_duty_cycle;
    buf[1] = hp_max;
    buf[2] = device_sel;
    buf[3] = pa_lut;
    __sx126x_write_command(p_this, RADIO_SET_PACONFIG, buf, 4);
}



static void __sx126x_set_dio_irq_params (radio_sx126x_fsk_dev_t *p_this,
                                         uint16_t                irq_mask,
                                         uint16_t                dio1_mask,
                                         uint16_t                dio2_mask,
                                         uint16_t                dio3_mask)
{
    uint8_t buf[8] = {0};

    buf[0] = (uint8_t)((irq_mask >> 8) & 0x00FF);
    buf[1] = (uint8_t)(irq_mask & 0x00FF);
    buf[2] = (uint8_t)((dio1_mask >> 8) & 0x00FF);
    buf[3] = (uint8_t)(dio1_mask & 0x00FF);
    buf[4] = (uint8_t)((dio2_mask >> 8) & 0x00FF);
    buf[5] = (uint8_t)(dio2_mask & 0x00FF);
    buf[6] = (uint8_t)((dio3_mask >> 8) & 0x00FF);
    buf[7] = (uint8_t)(dio3_mask & 0x00FF);
    __sx126x_write_command(p_this, RADIO_CFG_DIOIRQ, buf, 8);
}

static void __sx126x_fsk_addr_filter_set (radio_sx126x_fsk_dev_t *p_this, uint8_t addr_filter)
{
    uint8_t node_addr      = 0x55;
    uint8_t broadcast_addr = 0xAA;

    switch (addr_filter) {
    case FSK_ADDR_FILTER_DISABLE:
        break;

    case FSK_ADDR_FILTER_NODE:
        __sx126x_write_register(p_this, REG_ADDR_FSK_NODE_ADDR, node_addr);
        break;

    case FSK_ADDR_FILTER_NODE_BROADCAST:
        __sx126x_write_register(p_this, REG_ADDR_FSK_BROADCAST_ADDR, broadcast_addr);
        break;
    default:
        break;
    }
}


static int __sx126x_fsk_crc_seed_set (radio_sx126x_fsk_dev_t *p_this, uint8_t *p_crc_len)
{
    uint8_t  buf[2]        = {0};
    uint16_t crc_init_val  = 0;
    uint16_t crc_polynomia = 0;

    *p_crc_len = p_this->p_devinfo->p_fsk_param->crc_len;

    switch (p_this->p_devinfo->p_fsk_param->crc_len) {
    case FSK_CRC_2_BYTES_IBM:
        crc_init_val  = FSK_CRC_INIT_VAL_IBM;
        crc_polynomia = FSK_CRC_POLYNOMIAL_IBM;
        *p_crc_len    = FSK_CRC_2_BYTE;
        break;

    case FSK_CRC_2_BYTES_CCIT:
        crc_init_val  = FSK_CRC_INIT_VAL_CCIT;
        crc_polynomia = FSK_CRC_POLYNOMIAL_CCIT;
        *p_crc_len    = FSK_CRC_2_BYTE_INV;
        break;

    default:
        if (*p_crc_len > FSK_CRC_2_BYTES_CCIT) {
            return RADIO_RET_PARAM_INVALID;
        } else {
            return RADIO_RET_OK;
        }
    }

    buf[0] = (uint8_t)((crc_init_val >> 8) & 0xFF);
    buf[1] = (uint8_t)(crc_init_val & 0xFF);
    __sx126x_write_registers(p_this, REG_ADDR_FSK_CRC_INIT_MSB, buf, 2);

    buf[0] = (uint8_t)((crc_polynomia >> 8) & 0xFF);
    buf[1] = (uint8_t)(crc_polynomia & 0xFF);
    __sx126x_write_registers(p_this, REG_ADDR_FSK_CRC_POLY_MSB, buf, 2);

    return RADIO_RET_OK;
}

void __sx126x_fsk_whitening_seed_set (radio_sx126x_fsk_dev_t *p_this, uint8_t whitening_enable)
{
    uint8_t  reg_val = 0;
    uint16_t seed    = 0x01FF;

    if (whitening_enable == 0) {
        return;
    }

    reg_val = __sx126x_read_register(p_this, REG_ADDR_FSK_WHITEN_INIT_MSB);
    reg_val &= 0xFE;
    reg_val = ((seed >> 8) & 0x01) | reg_val;

    __sx126x_write_register(p_this, REG_ADDR_FSK_WHITEN_INIT_MSB, reg_val);
    __sx126x_write_register(p_this, REG_ADDR_FSK_WHITEN_INIT_LSB, (uint8_t)seed);
}


static int __sx126x_fsk_bandwitdh_cmd_get (uint32_t bandwidth_val, uint8_t *p_bandwidth_cmd)
{
    uint8_t total = sizeof(__g_sx126x_bandwidth) / sizeof(__g_sx126x_bandwidth[0]);

    for (uint8_t i = 0; i < total; i++) {
        if (bandwidth_val == __g_sx126x_bandwidth[i].bw_value) {
            *p_bandwidth_cmd = __g_sx126x_bandwidth[i].bw_cmd;
            return RADIO_RET_OK;
        }
    }

    return RADIO_RET_PARAM_INVALID;
}


static int __sx126x_fsk_set_modulation_params (radio_sx126x_fsk_dev_t *p_this, uint8_t rate_level)
{
    uint32_t data          = 0;
    uint8_t  buf[8]        = {0};
    uint32_t bit_rate      = 0;
    uint32_t freq_dev      = 0;
    uint32_t bandwidth     = 0;
    uint8_t  bandwidth_cmd = 0;
    uint8_t  filter_shape  = FSK_FILTER_GS_BT_1;

    if ((rate_level < 1) || (rate_level > 10)) {
        return RADIO_RET_PARAM_INVALID;
    }
    
    bit_rate  = __g_sx126x_rate_lv[rate_level].bit_rate;  
    freq_dev  = __g_sx126x_rate_lv[rate_level].freq_dev;   
    bandwidth = __g_sx126x_rate_lv[rate_level].bandwidth;  
    
    /* ֵתΪֵָ */
    if (__sx126x_fsk_bandwitdh_cmd_get(bandwidth, &bandwidth_cmd) != RADIO_RET_OK) {
        return RADIO_RET_PARAM_INVALID;
    }
    
    /* Ŀ,ȡдֵdata */
    data   = (uint32_t)((float)(32 * XTAL_FREQ) / (float)bit_rate);
    buf[0] = (data >> 16) & 0xFF;
    buf[1] = (data >> 8) & 0xFF;
    buf[2] = (data >> 0) & 0xFF;
    buf[3] = filter_shape;
    buf[4] = bandwidth_cmd;

    /* ĿƵƫȡдֵdata */
    data = ((float)(FREQ_DIV * freq_dev) / (float)XTAL_FREQ);
    buf[5] = (data >> 16) & 0xFF;
    buf[6] = (data >> 8) & 0xFF;
    buf[7] = (data >> 0) & 0xFF;

    __sx126x_write_command(p_this, RADIO_SET_MODULATIONPARAMS, buf, 8);
    return RADIO_RET_OK;
}



static int __sx126x_fsk_set_packet_params (radio_sx126x_fsk_dev_t *p_this)
{
    uint8_t  buf[9]           = {0};
    uint8_t  crc_len          = 0;
    uint16_t preamble_len     = p_this->p_devinfo->p_fsk_param->preamble_len;
    uint8_t  preamble_det_len = p_this->p_devinfo->p_fsk_param->preamble_det_len;
    uint8_t  addr_filter      = FSK_ADDR_FILTER_DISABLE;
    uint8_t  whitening_enable = 1;

    
    /* ǰ볤СֵΪ16bits */
    if (preamble_len < 16) {
        preamble_len = 16;
    }
    
    /* ǰ볤СֵΪ8bits */
    if (preamble_det_len < FSK_PREAMBLE_DETECTOR_BITS_8) {
        preamble_det_len = FSK_PREAMBLE_DETECTOR_BITS_8;
    }
        
    /* ǰ볤ֵΪ32bits */
    if (preamble_det_len > FSK_PREAMBLE_DETECTOR_BITS_32) {
        return RADIO_RET_PARAM_INVALID;;
    }
        

    /* ͬ볤 = ǰ * 2 */
    __g_syncword_len = (8 + (preamble_det_len - 4) * 8) * 2;

    if (__sx126x_fsk_crc_seed_set(p_this, &crc_len) != RADIO_RET_OK) {
        return RADIO_RET_PARAM_INVALID;
    }

    buf[0] = (preamble_len >> 8) & 0xFF;
    buf[1] = (preamble_len >> 0) & 0xFF;
    buf[2] = preamble_det_len;
    buf[3] = __g_syncword_len;
    buf[4] = addr_filter;
    buf[5] = !!p_this->p_devinfo->p_fsk_param->header_mode;

    if (SM_STATE_GET(p_this) == RFLR_STATE_TX_RUNNING) {
        buf[6] = p_this->p_devinfo->p_fsk_param->payload_len;
    } else if (SM_STATE_GET(p_this) == RFLR_STATE_RX_RUNNING) {
        buf[6] = RX_MAX_PAYLOAD_LENTH;
    } else {
    }

    buf[7] = crc_len;
    buf[8] = whitening_enable;

    __sx126x_write_command(p_this, RADIO_SET_PACKETPARAMS, buf, 9);

    uint8_t syncword[8] = {0xC1, 0x94, 0xC1, 0x02, 0x04, 0x06, 0x06, 0x04};
    __sx126x_sync_id_set(p_this, syncword, 8);

    __sx126x_fsk_addr_filter_set(p_this, addr_filter);

    __sx126x_fsk_whitening_seed_set(p_this, whitening_enable);

    return RADIO_RET_OK;
}


static void __sx126x_get_rx_buffer_status (radio_sx126x_fsk_dev_t *p_this,
                                           uint8_t                *p_payload_length,
                                           uint8_t                *p_rx_start_buf_ptr)
{
    uint8_t status[2] = {0};
    uint8_t reg_val   = 0;

    __sx126x_read_command(p_this, RADIO_GET_RXBUFFERSTATUS, status, 2);

    reg_val = __sx126x_read_register(p_this, REG_LR_PACKETPARAMS) >> 7;
    if (reg_val == 1) {
        *p_payload_length = __sx126x_read_register(p_this, REG_LR_PAYLOADLENGTH);
    } else {
        *p_payload_length = status[0];
    }
    *p_rx_start_buf_ptr = status[1];
}


static int __sx126x_set_sleep (void *p_drv)
{
    radio_sx126x_fsk_dev_t *p_this     = (radio_sx126x_fsk_dev_t *)p_drv;
    uint8_t                 value      = 0;
    uint8_t                 wakeup_rtc = 0;
    uint8_t                 reset      = 0;
    uint8_t                 warm_start = 1;
    uint8_t                 reserved   = 0;

    value = (reserved << 3) | (warm_start << 2) | (reset << 1) | (wakeup_rtc);
    __sx126x_write_command(p_this, RADIO_SET_SLEEP, &value, 1);
    __sx126x_delay_ms(p_this, 1);
    SM_STATE_SET(p_this, RFLR_STATE_SLEEP);

    return RADIO_RET_OK;
}


static int __sx126x_set_rx_mode (radio_sx126x_fsk_dev_t *p_this)
{
    uint8_t  buf[3]     = {0};
    uint32_t rx_timeout = 0;

    if (p_this->p_gpio->pfn_txen_pin_set) {
        p_this->p_gpio->pfn_txen_pin_set(0);
    }
    if (p_this->p_gpio->pfn_rxen_pin_set) {
        p_this->p_gpio->pfn_rxen_pin_set(1);
    }

    __sx126x_set_cap_val_of_xtal(p_this);

    __sx126x_set_stop_rxtimer_on_preamble_detect(p_this, false);

    SM_STATE_SET(p_this, RFLR_STATE_RX_RUNNING);

    __sx126x_fsk_set_packet_params(p_this);

    __sx126x_set_dio_irq_params(p_this,
                                IRQ_RX_DONE | IRQ_RX_TX_TIMEOUT | IRQ_CRC_ERROR,
                                IRQ_RX_DONE | IRQ_RX_TX_TIMEOUT | IRQ_CRC_ERROR,
                                IRQ_RADIO_NONE,
                                IRQ_RADIO_NONE);

    __sx126x_write_register(p_this, REG_RX_GAIN, 0x96);

    rx_timeout = p_this->p_devinfo->p_fsk_param->rx_pkt_timeout * 64;

    if (p_this->p_devinfo->p_fsk_param->rx_mode == 0) {
        rx_timeout = 0xFFFFFF;
    }

    buf[0] = (uint8_t)((rx_timeout >> 16) & 0xFF);
    buf[1] = (uint8_t)((rx_timeout >> 8) & 0xFF);
    buf[2] = (uint8_t)(rx_timeout & 0xFF);

    __sx126x_write_command(p_this, RADIO_SET_RX, buf, 3);

    return RADIO_RET_OK;
}



static int __sx126x_set_tx_continuous_wave (void *p_drv)
{
    radio_sx126x_fsk_dev_t *p_this = (radio_sx126x_fsk_dev_t *)p_drv;

    if (p_this->p_gpio->pfn_txen_pin_set) {
        p_this->p_gpio->pfn_txen_pin_set(1);
    }
    if (p_this->p_gpio->pfn_rxen_pin_set) {
        p_this->p_gpio->pfn_rxen_pin_set(0);
    }
    __sx126x_set_cap_val_of_xtal(p_this);

    __sx126x_freq_set(p_this, p_this->p_devinfo->p_fsk_param->frequency);
    __sx126x_set_rf_power(p_this, p_this->p_devinfo->p_fsk_param->power, RADIO_RAMP_800_US); 

    __sx126x_write_command(p_this, RADIO_SET_TXCONTINUOUSWAVE, 0, 0);
    SM_STATE_SET(p_this, RFLR_STATE_TX_RUNNING);
    return RADIO_RET_OK;
}


static int __sx126x_get_irq_status (void *p_drv)
{
    radio_sx126x_fsk_dev_t *p_this        = (radio_sx126x_fsk_dev_t *)p_drv;
    uint8_t                 irq_status[2] = {0};

    __sx126x_read_command(p_this, RADIO_GET_IRQSTATUS, irq_status, 2);

    return (irq_status[0] << 8) | irq_status[1];
}


static void __sx126x_clear_irq_status (void *p_drv, uint16_t irq)
{
    radio_sx126x_fsk_dev_t *p_this = (radio_sx126x_fsk_dev_t *)p_drv;
    uint8_t                 buf[2] = {0};

    buf[0] = (uint8_t)(((uint16_t)irq >> 8) & 0x00FF);
    buf[1] = (uint8_t)((uint16_t)irq & 0x00FF);
    __sx126x_write_command(p_this, RADIO_CLR_IRQSTATUS, buf, 2);
}



static int __sx126x_reset (void *p_arg)
{
    radio_sx126x_fsk_dev_t *p_this = (radio_sx126x_fsk_dev_t *)p_arg;
    if (p_this->p_gpio->pfn_reset_pin_set == NULL) {
        return -1;
    }

    __sx126x_delay_ms(p_this, 10);
    p_this->p_gpio->pfn_reset_pin_set(0);
    __sx126x_delay_ms(p_this, 20);
    p_this->p_gpio->pfn_reset_pin_set(1);
    __sx126x_delay_ms(p_this, 10);

    return RADIO_RET_OK;
}

static int __sx126x_sync_id_set (void *p_arg, uint8_t *p_id, uint8_t len)
{
    radio_sx126x_fsk_dev_t *p_this = (radio_sx126x_fsk_dev_t *)p_arg;

    if (len == 0) {
        return RADIO_RET_PARAM_INVALID;
    }

    __sx126x_write_registers(p_this, REG_ADDR_FSK_SYNCWORD0, p_id, (uint16_t)len);
    return RADIO_RET_OK;
}

static int __sx126x_sync_id_get (void *p_arg, uint8_t *p_id, uint8_t *p_len)
{
    radio_sx126x_fsk_dev_t *p_this = (radio_sx126x_fsk_dev_t *)p_arg;
    __sx126x_read_registers(p_this, REG_ADDR_FSK_SYNCWORD0, p_id, (uint16_t)*p_len);
    return RADIO_RET_OK;
}



static int __sx126x_preamble_length_set (void *p_arg, uint16_t preamble_len)
{
    radio_sx126x_fsk_dev_t *p_this = (radio_sx126x_fsk_dev_t *)p_arg;

    p_this->p_devinfo->p_fsk_param->preamble_len = preamble_len;
    if(preamble_len < 16) {
        p_this->p_devinfo->p_fsk_param->preamble_len = 16;
    }
    __sx126x_fsk_set_packet_params(p_this);
    return RADIO_RET_OK;
}



static int __sx126x_preamble_length_get (void *p_arg, uint16_t *p_len)
{
    radio_sx126x_fsk_dev_t *p_this = (radio_sx126x_fsk_dev_t *)p_arg;

    *p_len = p_this->p_devinfo->p_fsk_param->preamble_len;
    if(*p_len < 12) {
        *p_len = 12;
    }
    return RADIO_RET_OK;
}



static int __sx126x_set_rf_power (radio_sx126x_fsk_dev_t *p_this, 
                                  int8_t                  power, 
                                  uint8_t                 ramp_time)
{
    uint8_t  buf[2] = {0};
    uint8_t  pa_dc  = 0x04;   /* PAռձ Ĭ0x04 */
    uint8_t  hp_mx  = 0x07;   /* PAĴС Ĭ0x07 */
    
    __sx126x_set_pa_config(p_this, pa_dc, hp_mx, 0x00, 0x01); 

    if ((power < -9) || ((power > 22))) {
        return RADIO_RET_PARAM_INVALID;
    }
    
    __sx126x_write_register(p_this, REG_OCP_CONFIG, 0x38);

    buf[0] = power;
    buf[1] = ramp_time;

    p_this->p_devinfo->p_fsk_param->power = power;
    __sx126x_write_command(p_this, RADIO_SET_TXPARAMS, buf, 2);
    
    return RADIO_RET_OK;
}

static int __sx126x_power_set_adapter (void *p_arg, uint8_t level, float *p_power)
{
    uint8_t                 ret;
    radio_sx126x_fsk_dev_t *p_this = (radio_sx126x_fsk_dev_t *)p_arg;
    float power_tab[] = {-9, -5, -1, 3, 7, 12, 17, 22};  /* 8ʵȼ0 ~ 7 */

    if (level > (sizeof(power_tab) / sizeof(power_tab[0]))) {
        return RADIO_RET_CFG_OUT_OF_RANGE;
    }

    ret = __sx126x_set_rf_power(p_this, power_tab[level], RADIO_RAMP_800_US);
    if (ret == RADIO_RET_OK) {
        *p_power = p_this->p_devinfo->p_fsk_param->power = power_tab[level];
    } else {
        return RADIO_RET_PARAM_INVALID;
    }

    return RADIO_RET_OK;
}



static int __sx126x_freq_set (void *p_arg, uint32_t frequency)
{
    radio_sx126x_fsk_dev_t *p_this          = (radio_sx126x_fsk_dev_t *)p_arg;
    uint8_t                 buf[4]          = {0};
    uint32_t                freq            = 0;
    static bool             ImageCalibrated = false;

    if ((frequency < p_this->p_devinfo->p_fsk_param->dev_type.min_freq) ||
        (frequency > p_this->p_devinfo->p_fsk_param->dev_type.max_freq)) {
        return RADIO_RET_PARAM_INVALID;
    }

    if (ImageCalibrated == false) {
        __sx126x_calibrate_image(p_this, frequency);
        ImageCalibrated = true;
    }

    freq = (uint32_t)((double)frequency / (double)FREQ_STEP);
    buf[0] = (uint8_t)((freq >> 24) & 0xFF);
    buf[1] = (uint8_t)((freq >> 16) & 0xFF);
    buf[2] = (uint8_t)((freq >> 8) & 0xFF);
    buf[3] = (uint8_t)(freq & 0xFF);

    p_this->p_devinfo->p_fsk_param->frequency = frequency;
    __sx126x_write_command (p_this, RADIO_SET_RFFREQUENCY, buf, 4);
    return RADIO_RET_OK;
}



static int __sx126x_mode_set (void *p_arg, uint8_t mode)
{
    radio_sx126x_fsk_dev_t *p_this = (radio_sx126x_fsk_dev_t *)p_arg;
    int                     ret    = RADIO_RET_OK;

    switch (mode) {

    case SLEEP_MODE:
        ret = __sx126x_set_sleep(p_this);
        break;

    case STANDBY_MODE:
        ret = __sx126x_set_standby(p_this, STDBY_XOSC_MODE);
        break;

    case RX_MODE:
        ret = __sx126x_set_rx_mode(p_this);
        break;

    case CW_MODE:
        ret = __sx126x_set_tx_continuous_wave(p_this);
        break;

    case RX_BOOST_MODE:
        ret = __sx126x_radio_rx_boosted(p_this, 20000);
        break;

    case RX_DC_MODE:
        ret = __sx126x_set_rx_duty_cycle(p_this, 10000, 10000);
        break;

    case FS_MODE:
        ret = __sx126x_set_fs(p_this);
        break;

    default:
        break;
    }

    return ret;
}


static int __sx126x_state_get (void *p_arg, radio_state_t *p_state)
{
    radio_sx126x_fsk_dev_t *p_this = (radio_sx126x_fsk_dev_t *)p_arg;

    switch (SM_STATE_GET(p_this)) {
    case RFLR_STATE_TX_RUNNING:
    case RFLR_STATE_TX_INIT:
        *p_state = TX_RUNING_ST;
        break;

    case RFLR_STATE_TX_DONE:
    case RFLR_STATE_TX_TIMEOUT:
        *p_state = IDLE_ST;
        break;

    case RFLR_STATE_RX_DONE:
    case RFLR_STATE_RX_TIMEOUT:
        if (p_this->p_devinfo->p_fsk_param->rx_mode == CONTINUOUS_RECV_MODE) {
            *p_state = RX_RUNING_ST;
        } else {
            *p_state = IDLE_ST;
        }
        break;

    case RFLR_STATE_RX_INIT:
    case RFLR_STATE_RX_RUNNING:
        *p_state = RX_RUNING_ST;
        break;

    case RFLR_STATE_CAD_RUNNING:
    case RFLR_STATE_CAD_INIT:
        *p_state = CAD_ST;
        break;

    case RFLR_STATE_IDLE:
        *p_state = IDLE_ST;
        break;

    default:
        *p_state = IDLE_ST;
        break;
    }

    return RADIO_RET_OK;
}



static int __sx126x_cur_rssi_get (void *p_arg, int16_t *p_rssi)
{
    radio_sx126x_fsk_dev_t *p_this = (radio_sx126x_fsk_dev_t *)p_arg;
    uint8_t                 buf[1] = {0};

    __sx126x_read_command(p_this, RADIO_GET_RSSIINST, buf, 1);
    *p_rssi = (int16_t)(-buf[0] >> 1);

    return RADIO_RET_OK;
}



static int __sx126x_time_on_air_get (void     *p_arg,
                                     uint16_t  preamble_len,
                                     uint8_t   pkt_len,
                                     uint32_t *p_time)
{
    radio_sx126x_fsk_dev_t *p_this   = (radio_sx126x_fsk_dev_t *)p_arg;
    uint32_t                bit_rate = 0;
    
    bit_rate = __g_sx126x_rate_lv[p_this->p_devinfo->p_fsk_param->rate_level].bit_rate;
    
    if (preamble_len == RADIO_TIME_ON_AIR_USE_CUR_PREAMBLE_LEN) {
        preamble_len = p_this->p_devinfo->p_fsk_param->preamble_len;
    }

    *p_time = (uint32_t)rint((8 * ((preamble_len >> 3) +
              (__g_syncword_len >> 3) +
              ((p_this->p_devinfo->p_fsk_param->header_mode == 0) ? 0.0 : 1.0) +
              pkt_len +
              ((p_this->p_devinfo->p_fsk_param->crc_len == FSK_CRC_2_BYTE) ? 2.0 : 0)) /
              bit_rate) * 1000);

    return RADIO_RET_OK;
}


static int __sx126x_data_send (void *p_arg, uint8_t *p_buf, uint8_t size)
{
    radio_sx126x_fsk_dev_t *p_this       = (radio_sx126x_fsk_dev_t *)p_arg;
    uint8_t                 buf[3]       = {0};
    uint16_t                irq_regs     = 0;
    uint32_t                timeout      = 0;
    uint16_t                crc          = 0;
    uint8_t                 crc_bytes[2] = {0};
    uint32_t                time_on_air  = 0;

    if ((NULL == p_buf)) {
        return RADIO_RET_PARAM_INVALID;
    }

    if (p_this->p_gpio->pfn_txen_pin_set) {
        p_this->p_gpio->pfn_txen_pin_set(1);
    }
    if (p_this->p_gpio->pfn_rxen_pin_set) {
        p_this->p_gpio->pfn_rxen_pin_set(0);
    }

    if (p_this->p_devinfo->p_fsk_param->crc_software == 1) {
        if (size > 253) {
            return RADIO_RET_CFG_OUT_OF_RANGE;
        }
        crc = __crc16_ccitt(p_buf, size);
        crc_bytes[0] = (crc & 0xFF00) >> 8;
        crc_bytes[1] = crc & 0xFF;
        size += 2;
    } else {
        if (size > 255) {
            return RADIO_RET_CFG_OUT_OF_RANGE;
        }
    }

    __sx126x_set_cap_val_of_xtal(p_this);

    __sx126x_set_dio_irq_params(p_this,
                                IRQ_TX_DONE | IRQ_RX_TX_TIMEOUT,
                                IRQ_RADIO_NONE,
                                IRQ_RADIO_NONE,
                                IRQ_RADIO_NONE);

    p_this->p_devinfo->p_fsk_param->payload_len = size;

    SM_STATE_SET(p_this, RFLR_STATE_TX_RUNNING);

    __sx126x_fsk_set_packet_params(p_this);

    if (p_this->p_devinfo->p_fsk_param->crc_software == 1) {
        __sx126x_write_buffer(p_this, 0x00, p_buf, size - 2);
        __sx126x_write_buffer(p_this, 0x00 + size - 2, crc_bytes, 2);
    } else {
        __sx126x_write_buffer(p_this, 0x00, p_buf, size);
    }

    timeout = p_this->p_devinfo->p_fsk_param->tx_timeout;

    __sx126x_time_on_air_get(p_this,
                             RADIO_TIME_ON_AIR_USE_CUR_PREAMBLE_LEN,
                             size,
                             &time_on_air);
    if (timeout == 1) {
        timeout = (time_on_air + 20) * 64;
    } else {
        timeout = 0;
    }

    buf[0] = (uint8_t)((timeout >> 16) & 0xFF);
    buf[1] = (uint8_t)((timeout >> 8) & 0xFF);
    buf[2] = (uint8_t)(timeout & 0xFF);

    __sx126x_write_command(p_this, RADIO_SET_TX, buf, 3);

    while (1) {

        irq_regs = __sx126x_get_irq_status(p_this);
        if ((irq_regs & IRQ_TX_DONE) == IRQ_TX_DONE) {
            __sx126x_clear_irq_status(p_this, IRQ_RADIO_ALL);
            SM_STATE_SET(p_this, RFLR_STATE_TX_DONE);
            if (p_this->p_devinfo->p_events->pfn_tx_done_cb != NULL) {
                p_this->p_devinfo->p_events->pfn_tx_done_cb();
            }
            break;
        }

        if ((irq_regs & IRQ_RX_TX_TIMEOUT ) == IRQ_RX_TX_TIMEOUT) {
            __sx126x_clear_irq_status(p_this, IRQ_RADIO_ALL);
            SM_STATE_SET(p_this, RFLR_STATE_TX_TIMEOUT);
            if (p_this->p_devinfo->p_events->pfn_tx_timeout_cb != NULL) {
                p_this->p_devinfo->p_events->pfn_tx_timeout_cb();
            }
            return RADIO_RET_TX_TIMEOUT;
        }

    }

    return RADIO_RET_OK;
}




static int __sx126x_rx_state_check (void *p_arg, int *p_rx_state)
{
    radio_sx126x_fsk_dev_t *p_this   = (radio_sx126x_fsk_dev_t *)p_arg;
    uint16_t                irq_regs = 0;

    irq_regs = __sx126x_get_irq_status(p_this);

    if ((irq_regs & IRQ_RX_TX_TIMEOUT) == IRQ_RX_TX_TIMEOUT) {    /* ճʱ */
        if (p_this->p_devinfo->p_events->pfn_rx_timeout_cb != NULL) {
            p_this->p_devinfo->p_events->pfn_rx_timeout_cb();
        }
        *p_rx_state = RADIO_RET_RX_TIMEOUT;
    } else if ((irq_regs & IRQ_CRC_ERROR) == IRQ_CRC_ERROR) {    /* ӲCRC */
        if (p_this->p_devinfo->p_events->pfn_rx_hw_crc_error_cb != NULL) {
            p_this->p_devinfo->p_events->pfn_rx_hw_crc_error_cb();
        }
        *p_rx_state = RADIO_RET_RX_CRC_ERR;
    } else if ((irq_regs & IRQ_RX_DONE) == IRQ_RX_DONE) {    /*  */
        *p_rx_state = RADIO_RET_OK;
    } else {
        *p_rx_state = RADIO_RET_NOT_RECV_PKT;
    }

    /* ж֮Ŷжϱ־ */
    if (*p_rx_state != RADIO_RET_NOT_RECV_PKT) {
        __sx126x_clear_irq_status(p_this, IRQ_RADIO_ALL);
    }

    return RADIO_RET_OK;
}





static int __sx126x_data_recv (void *p_arg, uint8_t *p_buf, uint8_t *p_size)
{
    radio_sx126x_fsk_dev_t *p_this    = (radio_sx126x_fsk_dev_t *)p_arg;
    int                     rx_state  = 0;
    uint8_t                 offset    = 0;
    uint8_t                 status[3] = {0};
    int16_t                 pkt_rssi  = 0;
    int16_t                 pkt_snr   = 0;
    uint8_t                 pkt_size  = 0;
    uint16_t                crc       = 0;

    if (SM_STATE_GET(p_this) != RFLR_STATE_RX_RUNNING) {
        return RADIO_RET_NOT_RECV_PKT;
    }

    if ((p_buf == NULL) || (p_size == NULL)) {
        return RADIO_RET_PARAM_INVALID;
    }

    __sx126x_rx_state_check(p_this, &rx_state);

    if ((rx_state == RADIO_RET_RX_TIMEOUT) || (rx_state == RADIO_RET_RX_CRC_ERR)) {
        *p_size = 0;
        if (p_this->p_devinfo->p_fsk_param->rx_mode == CONTINUOUS_RECV_MODE) {
            SM_STATE_SET(p_this, RFLR_STATE_RX_RUNNING);
        } else {
            SM_STATE_SET(p_this, RFLR_STATE_IDLE);
        }
        return rx_state;
    } else if (rx_state == RADIO_RET_NOT_RECV_PKT) {
        SM_STATE_SET(p_this, RFLR_STATE_RX_RUNNING);
        return rx_state;
    } else {
        SM_STATE_SET(p_this, RFLR_STATE_RX_DONE);
    }

    __sx126x_get_rx_buffer_status(p_this, &pkt_size, &offset);

    if (p_this->p_devinfo->p_fsk_param->crc_software) {
        if (pkt_size < 3) {
            goto sw_crc_err;
        }
    } else {
        if (pkt_size < 1) {
            goto sw_crc_err;
        }
    }

    __sx126x_read_buffer(p_this, offset, p_buf, pkt_size);
    *p_size = pkt_size;

    if (p_this->p_devinfo->p_fsk_param->crc_software == 1) {
        crc = __crc16_ccitt(p_buf, *p_size - 2);
        if ((((crc & 0xFF00) >> 8) != p_buf[*p_size - 2]) ||
             ((crc & 0xFF) != p_buf[*p_size - 1])) {
            goto sw_crc_err;
        }
        *p_size -= 2;
    }

    __sx126x_read_command(p_this, RADIO_GET_PACKETSTATUS, status, 3);
    pkt_rssi  = -status[2] >> 1;

    if (p_this->p_devinfo->p_events != NULL) {
        if (p_this->p_devinfo->p_events->pfn_rx_done_cb != NULL) {
            p_this->p_devinfo->p_events->pfn_rx_done_cb(p_buf,
                                                        *p_size,
                                                        pkt_rssi,
                                                        pkt_snr);
        }
    }

    if (p_this->p_devinfo->p_fsk_param->rx_mode == CONTINUOUS_RECV_MODE) {
        SM_STATE_SET(p_this, RFLR_STATE_RX_RUNNING);
    } else {
        SM_STATE_SET(p_this, RFLR_STATE_IDLE);
    }

    return RADIO_RET_OK;

sw_crc_err:
    *p_size = 0;
    if (p_this->p_devinfo->p_events->pfn_rx_sw_crc_error_cb != NULL) {
        p_this->p_devinfo->p_events->pfn_rx_sw_crc_error_cb();
    }
    if (p_this->p_devinfo->p_fsk_param->rx_mode == CONTINUOUS_RECV_MODE) {
        SM_STATE_SET(p_this, RFLR_STATE_RX_RUNNING);
    } else {
        SM_STATE_SET(p_this, RFLR_STATE_IDLE);
    }
    return RADIO_RET_RX_CRC_ERR;
}

static int __sx126x_ioctl (void *p_arg, int cmd, void *p_param)
{
    radio_sx126x_fsk_dev_t *p_this     = (radio_sx126x_fsk_dev_t *)p_arg;
    int                     ret        = RADIO_RET_OK;
    uint8_t                 rate_level = 0;

    switch (cmd) {

    case RATE_LEVEL_SET:
        rate_level = *(uint8_t *)p_param;
        ret = __sx126x_fsk_set_modulation_params(p_this, rate_level);
        if (ret == RADIO_RET_OK) {
            p_this->p_devinfo->p_fsk_param->rate_level = rate_level;
        }
        break;

    default:
        ret = RADIO_RET_CMD_NOT_SUPPORT;
        break;
    }

    return ret;
}


static void __sx126x_dio1_irq_func (void *p_arg)
{
}


static const struct radio_drv_funcs __g_sx126x_drv_funcs =
{
    __sx126x_reset,
    __sx126x_sync_id_set,
    __sx126x_sync_id_get,
    __sx126x_preamble_length_set,
    __sx126x_preamble_length_get,
    __sx126x_power_set_adapter,
    __sx126x_freq_set,
    __sx126x_mode_set,
    __sx126x_state_get,
    __sx126x_cur_rssi_get,
    __sx126x_time_on_air_get,
    __sx126x_data_send,
    __sx126x_data_recv,
    __sx126x_ioctl,
    NULL,
    __sx126x_dio1_irq_func,
    NULL,
    NULL,
    NULL,
};



static int __sx126x_fsk_init (radio_sx126x_fsk_dev_t *p_this)
{
    int ret;

    SM_STATE_SET(p_this, RFLR_STATE_IDLE);

    ret = __sx126x_freq_set(p_this, p_this->p_devinfo->p_fsk_param->frequency);
    if (ret != RADIO_RET_OK) {
        return ret;
    }

    ret = __sx126x_set_rf_power(p_this, p_this->p_devinfo->p_fsk_param->power, RADIO_RAMP_800_US);
    if (ret != RADIO_RET_OK) {
        return ret;
    }

    __sx126x_mode_set(p_this, STANDBY_MODE);
    __sx126x_set_fsk_mode(p_this);

    ret = __sx126x_fsk_set_modulation_params(p_this, p_this->p_devinfo->p_fsk_param->rate_level);
    if (ret != RADIO_RET_OK) {
        return ret;
    }

    ret = __sx126x_fsk_set_packet_params(p_this);
    if (ret != RADIO_RET_OK) {
        return ret;
    }

    return RADIO_RET_OK;
}



radio_handle_t radio_sx126x_fsk_init (radio_sx126x_fsk_dev_t      *p_sx126x,
                                      sx126x_spi_funcs_t          *p_spi,
                                      sx126x_gpio_funcs_t         *p_gpio,
                                      sx126x_delay_funcs_t        *p_delay,
                                      sx126x_cpu_lock_funcs_t     *p_cpu_lock,
                                      radio_sx126x_fsk_devinfo_t  *p_devinfo)
{
    uint8_t calib_param = 0x7F;
    uint8_t test_ret    = 0;
    uint8_t val_tx_mod  = 0;
    uint8_t val_ant_mis = 0;

    if ((p_sx126x == NULL)||(p_spi == NULL)||(p_devinfo == NULL)) {
        return NULL;
    }

    p_sx126x->p_spi        = p_spi;
    p_sx126x->p_gpio       = p_gpio;
    p_sx126x->p_delay      = p_delay;
    p_sx126x->p_lock       = p_cpu_lock;
    p_sx126x->p_devinfo    = p_devinfo;
    p_sx126x->isa.p_funcs  = &__g_sx126x_drv_funcs;
    p_sx126x->isa.p_drv    = p_sx126x;

    SM_STATE_SET(p_sx126x, RFLR_STATE_IDLE);

    __sx126x_reset(p_sx126x);

    __sx126x_wake_up(p_sx126x);
    __sx126x_set_standby(p_sx126x, STDBY_RC_MODE);

    if (p_sx126x->p_devinfo->p_fsk_param->dev_type.osc_type == OSC_TCXO) {
        __sx126x_clear_device_errors(p_sx126x);
        __sx126x_set_dio3_as_tcxo_ctrl(p_sx126x, TCXO_CTRL_3_3V, 100);
        __sx126x_calibrate(p_sx126x, calib_param);
    }

    __sx126x_write_register(p_sx126x, REG_LR_PAYLOADLENGTH, 0x11);
    test_ret = __sx126x_read_register(p_sx126x, REG_LR_PAYLOADLENGTH);
    if (test_ret != 0x11) {
        return NULL;
    }

    __sx126x_set_dio2_as_rf_switch_ctrl(p_sx126x, true);
    __sx126x_set_standby(p_sx126x, STDBY_RC_MODE);
    __sx126x_set_regulator_mode(p_sx126x, USE_DCDC);
    __sx126x_set_buffer_base_address(p_sx126x, 0x00, 0x00);

    if (__sx126x_fsk_init(p_sx126x) != RADIO_RET_OK) {
        return NULL;
    }

    val_tx_mod = __sx126x_read_register(p_sx126x, REG_TX_MODULATION);
    val_tx_mod = val_tx_mod | 0x04;
    __sx126x_write_register(p_sx126x, REG_TX_MODULATION, val_tx_mod);

    val_ant_mis = __sx126x_read_register(p_sx126x, REG_TX_CLAMP_CONFIG);
    val_ant_mis = val_ant_mis | 0x1E;
    __sx126x_write_register(p_sx126x, REG_TX_CLAMP_CONFIG, val_ant_mis);

    return &p_sx126x->isa;
}



