/*******************************************************************************
*                                 ZM4xx
*                       ----------------------------
*                       innovating embedded platform
*
* Copyright (c) 2001-present Guangzhou ZHIYUAN Electronics Co., Ltd.
* ALL rights reserved.
*
* THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
* IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
* FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. 
* 
* The software is free to use in a commercial product and should not be
* redistributed in any way. See the LICENSE file for more details.
* 
* Contact information:
* web site:    http://www.zlg.cn/
*******************************************************************************/

/**
 * \file
 * \brief  SX126x LoRa driver
 *
 * \internal
 * \par Modification history
 * - v1.3.0 25-05-13
 * - v1.2.9 24-04-10
 * - v1.2.8 23-12-24
 * - v1.2.7 23-08-24
 * - v1.2.6 23-05-23
 * - v1.2.5 22-12-27
 * - v1.2.4 22-11-17
 * - v1.2.3 22-08-24
 * - v1.2.2 22-08-22
 * - v1.1.6 22-05-16
 * - v1.1.3 21-09-15
 * - v1.1.1 21-08-03
 * - v1.1.0 21-01-26
 * - v1.0.0 20-09-08  sup, first implementation.
 * \endinternal
 */


#include <math.h>
#include "string.h"
#include "stdio.h"
#include "sx126x_lora.h"
 

#define SM_STATE_SET(p_this, state)   ((p_this)->run_state.sm_state = state)
#define SM_STATE_GET(p_this)          ((p_this)->run_state.sm_state)

#define CAD_STATE_SET(p_this, state)  ((p_this)->run_state.cad_state |= state)
#define CAD_STATE_GET(p_this)         ((p_this)->run_state.cad_state)
#define CAD_STATE_CLR(p_this)         ((p_this)->run_state.cad_state = 0)



static int  __sx126x_mode_set (void *p_arg, uint8_t mode);
static int  __sx126x_cur_rssi_get (void *p_arg, int16_t *p_rssi);
static int  __sx126x_freq_set (void *p_arg, uint32_t frequency);
static int  __sx126x_set_rf_power (radio_sx126x_lora_dev_t *p_this, 
                                   int8_t                   power, 
                                   uint8_t                  ramp_time);
static void __sx126x_set_rx (radio_sx126x_lora_dev_t *p_this, uint32_t timeout); 
static void __sx126x_set_dio_irq_params (radio_sx126x_lora_dev_t *p_this, 
                                         uint16_t                 irq_mask, 
                                         uint16_t                 dio1_mask, 
                                         uint16_t                 dio2_mask, 
                                         uint16_t                 dio3_mask);
static int __sx126x_set_standby (radio_sx126x_lora_dev_t *p_this, uint8_t standby_mode);
static void __sx126x_set_lora_mode (void *p_drv);
static int  __sx126x_set_cad (radio_sx126x_lora_dev_t *p_this, 
                              uint8_t                  cad_mode, 
                              uint32_t                 rx_timeout);
static int  __sx126x_cad_start (radio_sx126x_lora_dev_t *p_this, uint8_t *p_state);


static const uint8_t __g_sx126x_band_width[] = {LORA_BW_007,
                                                LORA_BW_010,
                                                LORA_BW_015,
                                                LORA_BW_020,
                                                LORA_BW_031,
                                                LORA_BW_041,
                                                LORA_BW_062,
                                                LORA_BW_125,
                                                LORA_BW_250,
                                                LORA_BW_500}; 


static void __sx126x_delay_ms (radio_sx126x_lora_dev_t *p_this, uint32_t ms)
{
    p_this->p_delay->pfn_delay_ms(ms);
}

static void __sx126x_delay_us (radio_sx126x_lora_dev_t *p_this, uint32_t us)
{
    p_this->p_delay->pfn_delay_us(us);
}

static void __send_byte (radio_sx126x_lora_dev_t *p_this, uint8_t byte)
{
    p_this->p_spi->pfn_spi_write_byte(byte);
}

static uint8_t __recv_byte (radio_sx126x_lora_dev_t *p_this)
{
    return p_this->p_spi->pfn_spi_read_byte();
}


static void __sx126x_wait_on_busy (radio_sx126x_lora_dev_t *p_this)
{
    while(1) {
        if (p_this->p_gpio->pfn_busy_pin_read() == 0) {  
            break;
        } 
    }
}


static void __sx126x_wake_up (radio_sx126x_lora_dev_t *p_this)
{
    uint32_t key;
    uint8_t  status   = 0;
    uint8_t  cpu_busy = 0;

    key = p_this->p_lock->pfn_cpu_lock();

    do {
        p_this->p_gpio->pfn_sel_pin_set(0);
        __send_byte(p_this, RADIO_GET_STATUS);
        __send_byte(p_this, 0x00);
        status = __recv_byte(p_this);
        cpu_busy = (status & 0x80) >> 7;
        p_this->p_gpio->pfn_sel_pin_set(1);
    } while (cpu_busy == 1);

    p_this->p_lock->pfn_cpu_unlock(key);
    __sx126x_wait_on_busy(p_this);
}


static void __sx126x_check_device_ready (radio_sx126x_lora_dev_t *p_this)
{
    if ((SM_STATE_GET(p_this) == RFLR_STATE_SLEEP) || 
        (SM_STATE_GET(p_this) == RFLR_STATE_RX_DC)) {
        __sx126x_wake_up(p_this);
    }
    __sx126x_wait_on_busy(p_this);
}


static void __sx126x_write_command (radio_sx126x_lora_dev_t *p_this, 
                                    radio_commands_t         command, 
                                    uint8_t                 *p_buffer, 
                                    uint16_t                 size)
{
    uint16_t i;
    uint32_t key;
    
    __sx126x_check_device_ready(p_this);
    
    key = p_this->p_lock->pfn_cpu_lock();
    p_this->p_gpio->pfn_sel_pin_set(0);
    __send_byte(p_this, (uint8_t)command);

    for (i = 0; i < size; i++) {
        __send_byte(p_this, p_buffer[i]);
    }

    p_this->p_gpio->pfn_sel_pin_set(1);
    p_this->p_lock->pfn_cpu_unlock(key);
    
    if (command != RADIO_SET_SLEEP) {
        __sx126x_wait_on_busy(p_this);
    }
}

static void __sx126x_read_command (radio_sx126x_lora_dev_t *p_this, 
                                   radio_commands_t         command, 
                                   uint8_t                 *p_buffer, 
                                   uint16_t                 size)
{
    uint16_t i;
    uint32_t key;

    __sx126x_check_device_ready(p_this);
    
    key = p_this->p_lock->pfn_cpu_lock();
    p_this->p_gpio->pfn_sel_pin_set(0);
    
    __send_byte(p_this, (uint8_t)command);
    __send_byte(p_this, 0x00);

    for (i = 0; i < size; i++) {
        p_buffer[i] = __recv_byte(p_this);
    }

    p_this->p_gpio->pfn_sel_pin_set(1);
    p_this->p_lock->pfn_cpu_unlock(key);
    
    __sx126x_wait_on_busy(p_this);
}

static void __sx126x_write_registers (radio_sx126x_lora_dev_t *p_this, 
                                      uint16_t                 address, 
                                      uint8_t                 *p_buffer, 
                                      uint16_t                 size)
{
    uint16_t i;
    uint32_t key;
    
    __sx126x_check_device_ready(p_this);
    
    key = p_this->p_lock->pfn_cpu_lock();
    p_this->p_gpio->pfn_sel_pin_set(0);
    
    __send_byte(p_this, RADIO_WRITE_REGISTER);
    __send_byte(p_this, (address & 0xFF00) >> 8);
    __send_byte(p_this, address & 0x00FF);

    for (i = 0; i < size; i++) {
        __send_byte(p_this, p_buffer[i]);
    }

    p_this->p_gpio->pfn_sel_pin_set(1);
    p_this->p_lock->pfn_cpu_unlock(key);
    
    __sx126x_wait_on_busy(p_this);
}

static void __sx126x_write_register (radio_sx126x_lora_dev_t *p_this, 
                                     uint16_t                 address, 
                                     uint8_t                  value)
{
    __sx126x_write_registers(p_this, address, &value, 1);
}

static void __sx126x_read_registers (radio_sx126x_lora_dev_t *p_this, 
                                     uint16_t                 address, 
                                     uint8_t                 *p_buffer, 
                                     uint16_t                 size)
{
    uint16_t i;
    uint32_t key;
    
    __sx126x_check_device_ready(p_this);
    
    key = p_this->p_lock->pfn_cpu_lock();
    p_this->p_gpio->pfn_sel_pin_set(0);

    __send_byte(p_this, RADIO_READ_REGISTER);
    __send_byte(p_this, (address & 0xFF00) >> 8);
    __send_byte(p_this, address & 0x00FF);
    __send_byte(p_this, 0x00);

    for (i = 0; i < size; i++) {
        p_buffer[i] = __recv_byte(p_this);
    }
    p_this->p_gpio->pfn_sel_pin_set(1);
    p_this->p_lock->pfn_cpu_unlock(key);

    __sx126x_wait_on_busy(p_this);
}

static uint8_t __sx126x_read_register (radio_sx126x_lora_dev_t *p_this, 
                                       uint16_t                 address)
{
    uint8_t data;
    __sx126x_read_registers(p_this, address, &data, 1);
    return data;
}

static void __sx126x_write_buffer (radio_sx126x_lora_dev_t *p_this, 
                                   uint8_t                  offset, 
                                   uint8_t                 *p_buffer, 
                                   uint8_t                  size)
{
    uint16_t i;
    uint32_t key;
    
    __sx126x_check_device_ready(p_this);
    
    key = p_this->p_lock->pfn_cpu_lock();
    p_this->p_gpio->pfn_sel_pin_set(0);

    __send_byte(p_this, RADIO_WRITE_BUFFER);
    __send_byte(p_this, offset);

    for (i = 0; i < size; i++) {
      __send_byte(p_this, p_buffer[i]);
    }

    p_this->p_gpio->pfn_sel_pin_set(1);
    p_this->p_lock->pfn_cpu_unlock(key);
    
    __sx126x_wait_on_busy(p_this);
}

static void __sx126x_read_buffer (radio_sx126x_lora_dev_t *p_this, 
                                  uint8_t                  offset, 
                                  uint8_t                 *p_buffer, 
                                  uint8_t                  size)
{
    uint16_t i;
    uint32_t key;
    
    __sx126x_check_device_ready(p_this);
    
    key = p_this->p_lock->pfn_cpu_lock();
    p_this->p_gpio->pfn_sel_pin_set(0);

    __send_byte(p_this, RADIO_READ_BUFFER);
    __send_byte(p_this, offset);
    __send_byte(p_this, 0x00);

    for (i = 0; i < size; i++) {
        p_buffer[i] = __recv_byte(p_this);
    }
    
    p_this->p_gpio->pfn_sel_pin_set(1);
    p_this->p_lock->pfn_cpu_unlock(key);
    
    __sx126x_wait_on_busy(p_this);
}


static uint16_t __crc16_ccitt (uint8_t *p_buf, uint32_t len)
{
    uint16_t ccitt16 = 0x1021;
    uint16_t crc     = 0;
    int      i, j;
    for (i = 0; i < len; i++) {
        crc ^= (p_buf[i] << 8);
        for (j = 0; j < 8; j++) {
            if (crc & 0x8000) {
                crc <<= 1;
                crc ^= ccitt16;
            } else {
                crc <<= 1;
            }
        }
    }
    return crc;
}



static int __sx126x_set_cap_val_of_xtal (radio_sx126x_lora_dev_t *p_this)
{
    uint8_t cap_val = p_this->p_devinfo->p_lora_param->dev_type.cap_val;
    if (cap_val > 47) {
        return RADIO_RET_CFG_OUT_OF_RANGE;
    }
    __sx126x_set_standby(p_this, STDBY_XOSC_MODE);
    __sx126x_write_register(p_this, REG_XTA_TRIM, cap_val);
    __sx126x_write_register(p_this, REG_XTB_TRIM, cap_val);
    return RADIO_RET_OK;
}

static int __sx126x_set_fs (radio_sx126x_lora_dev_t *p_this)
{
    __sx126x_write_command(p_this, RADIO_SET_FS, 0, 0);
    SM_STATE_SET(p_this, RFLR_STATE_IDLE);
    return RADIO_RET_OK;
}


static uint32_t __sx126x_get_random (radio_sx126x_lora_dev_t *p_this)  
{
    uint8_t buf[4] = {0};

    __sx126x_set_rx(p_this, 0xFFFFFF);  
    __sx126x_delay_ms(p_this, 1);
    __sx126x_read_registers(p_this, REG_RAND_NUM_GEN, buf, 4);
    __sx126x_set_standby(p_this, STDBY_XOSC_MODE);

    return (buf[0] << 24) | (buf[1] << 16) | (buf[2] << 8) | buf[3];
}




static int __sx126x_set_rx_duty_cycle (radio_sx126x_lora_dev_t *p_this, 
                                       uint32_t                 rx_time, 
                                       uint32_t                 sleep_time)
{
    uint8_t buf[6] = {0};

    buf[0] = (uint8_t)((rx_time >> 16) & 0xFF);
    buf[1] = (uint8_t)((rx_time >> 8) & 0xFF);
    buf[2] = (uint8_t)(rx_time & 0xFF);
    buf[3] = (uint8_t)((sleep_time >> 16) & 0xFF);
    buf[4] = (uint8_t)((sleep_time >> 8) & 0xFF);
    buf[5] = (uint8_t)(sleep_time & 0xFF);
    __sx126x_write_command(p_this, RADIO_SET_RXDUTYCYCLE, buf, 6);
    SM_STATE_SET(p_this, RFLR_STATE_RX_RUNNING);  
    
    return RADIO_RET_OK;
}


static void __sx126x_set_tx_infinite_preamble (radio_sx126x_lora_dev_t *p_this)
{
    __sx126x_write_command(p_this, RADIO_SET_TXCONTINUOUSPREAMBLE, 0, 0);
}


static void __sx126x_set_rx_tx_fallback_mode (radio_sx126x_lora_dev_t *p_this, 
                                              uint8_t                  fallback_mode)
{
    __sx126x_write_command(p_this, RADIO_SET_TXFALLBACKMODE, &fallback_mode, 1);
}



static void __sx126x_get_status (radio_sx126x_lora_dev_t *p_this)
{
    uint8_t status  = 0;  
    __sx126x_read_command(p_this, RADIO_GET_STATUS, (uint8_t *)&status, 1);
}


static uint16_t __sx126x_get_device_errors (radio_sx126x_lora_dev_t *p_this)
{
    uint16_t error_val;
    __sx126x_read_command(p_this, RADIO_GET_ERROR, (uint8_t *)&error_val, 2);
    return error_val;
}



static void __sx126x_clear_device_errors (radio_sx126x_lora_dev_t *p_this)
{
    uint8_t buf[2] = {0};
    __sx126x_write_command(p_this, RADIO_CLR_ERROR, buf, 2);
}



static uint32_t __sx126x_radio_random (void *p_drv)
{
    radio_sx126x_lora_dev_t *p_this = (radio_sx126x_lora_dev_t *)p_drv;
    uint8_t                  i      = 0;
    uint32_t                 rnd    = 0;
    int16_t                  rssi   = 0;

    __sx126x_set_lora_mode(p_this);
    __sx126x_set_rx(p_this, 0xFFFFFF);

    for (i = 0; i < 32; i++) {
        __sx126x_delay_ms(p_this, 1);
        __sx126x_cur_rssi_get(p_this, &rssi);
        rnd |= ((uint32_t)rssi & 0x01) << i; 
    }

    __sx126x_mode_set(p_this, SLEEP_MODE);
    return rnd;
}



static void __sx126x_set_rx_boosted (radio_sx126x_lora_dev_t *p_this, uint32_t timeout)
{
    uint8_t buf[3] = {0};

    __sx126x_write_register(p_this, REG_RX_GAIN, 0x96);  

    buf[0] = (uint8_t)((timeout >> 16) & 0xFF);
    buf[1] = (uint8_t)((timeout >> 8) & 0xFF);
    buf[2] = (uint8_t)(timeout & 0xFF);
    __sx126x_write_command(p_this, RADIO_SET_RX, buf, 3);
    SM_STATE_SET(p_this, RFLR_STATE_RX_RUNNING);
}


static int __sx126x_radio_rx_boosted (void *p_drv, uint32_t timeout)
{
    radio_sx126x_lora_dev_t *p_this = (radio_sx126x_lora_dev_t *)p_drv;
    
    __sx126x_set_dio_irq_params(p_this, 
                                IRQ_RADIO_ALL,   
                                IRQ_RADIO_ALL, 
                                IRQ_RADIO_NONE,
                                IRQ_RADIO_NONE);

    if (p_this->p_devinfo->p_lora_param->rx_mode == 0) {
        __sx126x_set_rx_boosted(p_this, 0xFFFFFF);  
    } else {
        __sx126x_set_rx_boosted(p_this, timeout << 6);
    }
    
    return RADIO_RET_OK;
}



static void __sx126x_set_public_network (void *p_drv, uint8_t enable)
{
    radio_sx126x_lora_dev_t *p_this           = (radio_sx126x_lora_dev_t *)p_drv;
    uint16_t                 private_syncword = 0x1424;
    uint16_t                 public_syncword  = 0x3444;

    if (enable == 1) {   
        __sx126x_write_register(p_this,
                                REG_ADDR_LORA_SYNCWORD_MSB,   
                                (public_syncword >> 8) & 0xFF);
        __sx126x_write_register(p_this, 
                                REG_ADDR_LORA_SYNCWORD_LSB,   
                                public_syncword & 0xFF);
    } else {   
        __sx126x_write_register(p_this, 
                                REG_ADDR_LORA_SYNCWORD_MSB,   
                                (private_syncword >> 8) & 0xFF);
        __sx126x_write_register(p_this, 
                                REG_ADDR_LORA_SYNCWORD_LSB,   
                                private_syncword & 0xFF);
    }
}



static void __sx126x_set_stop_rxtimer_on_preamble_detect (radio_sx126x_lora_dev_t *p_this, 
                                                          bool                     enable)
{
    __sx126x_write_command(p_this, RADIO_SET_STOPRXTIMERONPREAMBLE, (uint8_t*)&enable, 1);
}

static void __sx126x_set_lora_symb_num_timeout (radio_sx126x_lora_dev_t *p_this, 
                                                uint8_t                  symb_num)
{
    __sx126x_write_command(p_this, RADIO_SET_LORASYMBTIMEOUT, &symb_num, 1);
}

static void __sx126x_calibrate (radio_sx126x_lora_dev_t *p_this, uint8_t calib_param)
{
    __sx126x_write_command(p_this, RADIO_CALIBRATE, (uint8_t *)&calib_param, 1);
}

static void __sx126x_set_regulator_mode (radio_sx126x_lora_dev_t *p_this, uint8_t mode)
{
    __sx126x_write_command(p_this, RADIO_SET_REGULATORMODE, (uint8_t *)&mode, 1);
}

static void __sx126x_set_packet_type (radio_sx126x_lora_dev_t *p_this, uint8_t packet_type)
{
    __sx126x_write_command(p_this, RADIO_SET_PACKETTYPE, (uint8_t *)&packet_type, 1);
}

static uint8_t __sx126x_get_packet_type (radio_sx126x_lora_dev_t *p_this)
{
    uint8_t packet_type = 0;
    __sx126x_read_command(p_this, RADIO_GET_PACKETTYPE, (uint8_t *)&packet_type, 1);
    return packet_type;
}

static void __sx126x_calibrate_image (radio_sx126x_lora_dev_t *p_this, uint32_t freq)
{
    uint8_t cal_freq[2] = {0};

    if (freq > 900000000) {
        cal_freq[0] = 0xE1;
        cal_freq[1] = 0xE9;
    } else if (freq > 850000000) {
        cal_freq[0] = 0xD7;
        cal_freq[1] = 0xDB;
    } else if (freq > 770000000) {
        cal_freq[0] = 0xC1;
        cal_freq[1] = 0xC5;
    } else if (freq > 460000000) {
        cal_freq[0] = 0x75;
        cal_freq[1] = 0x81;
    } else if (freq > 425000000) {
        cal_freq[0] = 0x6B;
        cal_freq[1] = 0x6F;
    }

    __sx126x_write_command (p_this, RADIO_CALIBRATEIMAGE, cal_freq, 2);
}

static void __sx126x_set_dio2_as_rf_switch_ctrl (radio_sx126x_lora_dev_t *p_this, 
                                                 uint8_t                  enable)
{
    __sx126x_write_command(p_this, RADIO_SET_RFSWITCHMODE, &enable, 1);
}

static void __sx126x_set_dio3_as_tcxo_ctrl (radio_sx126x_lora_dev_t *p_this, 
                                            uint8_t                  tcxo_voltage, 
                                            uint32_t                 timeout)
{
    uint8_t buf[4] = {0};

    buf[0] = tcxo_voltage & 0x07;
    buf[1] = (uint8_t)((timeout >> 16) & 0xFF);
    buf[2] = (uint8_t)((timeout >> 8) & 0xFF);
    buf[3] = (uint8_t)(timeout & 0xFF);

    __sx126x_write_command(p_this, RADIO_SET_TCXOMODE, buf, 4);
}

static void __sx126x_set_buffer_base_address (radio_sx126x_lora_dev_t *p_this, 
                                              uint8_t                  tx_base_address, 
                                              uint8_t                  rx_base_address)
{
    uint8_t buf[2] = {0};

    buf[0] = tx_base_address;
    buf[1] = rx_base_address;
    __sx126x_write_command(p_this, RADIO_SET_BUFFERBASEADDRESS, buf, 2);
}

static void __sx126x_set_lora_mode (void *p_drv)
{
    radio_sx126x_lora_dev_t *p_this = (radio_sx126x_lora_dev_t *)p_drv;
    __sx126x_set_packet_type(p_this, PACKET_TYPE_LORA);
    __sx126x_set_public_network(p_this, p_this->p_devinfo->p_lora_param->public_net_enable);
}

static int __sx126x_set_standby (radio_sx126x_lora_dev_t *p_this, uint8_t standby_mode) 
{
    __sx126x_write_command(p_this, RADIO_SET_STANDBY, (uint8_t*)&standby_mode, 1);
    SM_STATE_SET(p_this, RFLR_STATE_IDLE);
    return RADIO_RET_OK;
}



static void __sx126x_set_rx (radio_sx126x_lora_dev_t *p_this, uint32_t timeout)  
{
    uint8_t buf[3] = {0};

    buf[0] = (uint8_t)((timeout >> 16) & 0xFF);
    buf[1] = (uint8_t)((timeout >> 8) & 0xFF);
    buf[2] = (uint8_t)(timeout & 0xFF);
    __sx126x_write_command(p_this, RADIO_SET_RX, buf, 3);
    SM_STATE_SET(p_this, RFLR_STATE_RX_RUNNING);
}



static void __sx126x_set_pa_config (radio_sx126x_lora_dev_t *p_this, 
                                    uint8_t                  pa_duty_cycle, 
                                    uint8_t                  hp_max, 
                                    uint8_t                  device_sel, 
                                    uint8_t                  pa_lut)
{
    uint8_t buf[4] = {0};

    buf[0] = pa_duty_cycle;
    buf[1] = hp_max;
    buf[2] = device_sel;
    buf[3] = pa_lut;
    __sx126x_write_command(p_this, RADIO_SET_PACONFIG, buf, 4);
}



static void __sx126x_set_dio_irq_params (radio_sx126x_lora_dev_t *p_this, 
                                         uint16_t                 irq_mask, 
                                         uint16_t                 dio1_mask, 
                                         uint16_t                 dio2_mask, 
                                         uint16_t                 dio3_mask)
{
    uint8_t buf[8] = {0};

    buf[0] = (uint8_t)((irq_mask >> 8) & 0x00FF);
    buf[1] = (uint8_t)(irq_mask & 0x00FF);
    buf[2] = (uint8_t)((dio1_mask >> 8) & 0x00FF);
    buf[3] = (uint8_t)(dio1_mask & 0x00FF);
    buf[4] = (uint8_t)((dio2_mask >> 8) & 0x00FF);
    buf[5] = (uint8_t)(dio2_mask & 0x00FF);
    buf[6] = (uint8_t)((dio3_mask >> 8) & 0x00FF);
    buf[7] = (uint8_t)(dio3_mask & 0x00FF);
    __sx126x_write_command(p_this, RADIO_CFG_DIOIRQ, buf, 8);
}



static int __sx126x_lora_set_modulation_params (radio_sx126x_lora_dev_t *p_this, 
                                                uint8_t                  sf, 
                                                uint8_t                  bw, 
                                                uint8_t                  cr)
{
    uint8_t ldro       = 0;
    uint8_t buf[4]     = {0};
    uint8_t val_tx_mod = 0;
    
    if (p_this->p_devinfo->p_lora_param->dev_type.rf_type == RADIO_SX126X) {
        if ((bw < 6) || (bw > 9)) {
            return RADIO_RET_PARAM_INVALID;
        } 
        if ((sf < 5) || (sf > 12)) {
            return RADIO_RET_PARAM_INVALID;
        }
    } else if (p_this->p_devinfo->p_lora_param->dev_type.rf_type == RADIO_LLCC68) {
        if ((bw < 7) || (bw > 9)) {
            return RADIO_RET_PARAM_INVALID;
        }
        
        if (sf < 5) {
            return RADIO_RET_PARAM_INVALID;
        }
        if ((bw == 7) && (sf > 9)) {
            return RADIO_RET_PARAM_INVALID;
        } 
        if ((bw == 8) && (sf > 10)) {
            return RADIO_RET_PARAM_INVALID;
        } 
        if ((bw == 9) && (sf > 11)) {
            return RADIO_RET_PARAM_INVALID;
        }
    } else {
        return RADIO_RET_PARAM_INVALID;
    }

    if ((cr < 1) || (cr > 4)) {
        return RADIO_RET_PARAM_INVALID;
    }

    if (p_this->p_devinfo->p_lora_param->ldro_auto == 1) {
        if(((bw == 6) && ((sf == 10) || (sf == 11)   || (sf == 12))) ||
           ((bw == 7) && ((sf == 11) || (sf == 12))) ||
           ((bw == 8) &&  (sf == 12))) {
            ldro = 0x01;
        } else {
            ldro = 0x00;
        }
    } else {
        ldro = !!p_this->p_devinfo->p_lora_param->ldro_enable;
    }

    buf[0] = sf;
    buf[1] = __g_sx126x_band_width[bw];
    buf[2] = cr;
    buf[3] = ldro;
    
    __sx126x_write_command(p_this, RADIO_SET_MODULATIONPARAMS, buf, 4);
    
    if (bw == 9) {
        val_tx_mod = __sx126x_read_register(p_this, REG_TX_MODULATION);
        val_tx_mod = val_tx_mod & 0xFB;
        __sx126x_write_register(p_this, REG_TX_MODULATION, val_tx_mod);
    } else {
        val_tx_mod = __sx126x_read_register(p_this, REG_TX_MODULATION);
        val_tx_mod = val_tx_mod | 0x04;
        __sx126x_write_register(p_this, REG_TX_MODULATION, val_tx_mod);
    }
    
    return RADIO_RET_OK;
}

static void __sx126x_lora_set_packet_params (radio_sx126x_lora_dev_t *p_this)
{
    uint32_t preamble_len = p_this->p_devinfo->p_lora_param->preamble_len; 
    uint8_t  sf           = p_this->p_devinfo->p_lora_param->sf;
    uint8_t  head_mode    = !!p_this->p_devinfo->p_lora_param->header_mode;
    uint8_t  buf[6]       = {0};

    if ((sf == 5) || (sf == 6)) {
        if (p_this->p_devinfo->p_lora_param->preamble_len < 12) {
            preamble_len = 12;
        }
    } 
    
    buf[0] = (preamble_len >> 8) & 0xFF;
    buf[1] = preamble_len;
    buf[2] = head_mode;

    if (SM_STATE_GET(p_this) == RFLR_STATE_TX_RUNNING) {
        buf[3] = p_this->p_devinfo->p_lora_param->payload_len;
    } else if (SM_STATE_GET(p_this) == RFLR_STATE_RX_RUNNING) {
        if (head_mode == 0) {  /* Ϊʽͷ */
            buf[3] = RX_MAX_PAYLOAD_LENTH;
        } else {  /* Ϊʽͷ */
            buf[3] = p_this->p_devinfo->p_lora_param->payload_len;
        }
    } else {
    }

    buf[4] = !!p_this->p_devinfo->p_lora_param->crc_on;
    buf[5] = !!p_this->p_devinfo->p_lora_param->invert_iq;

    __sx126x_write_command(p_this, RADIO_SET_PACKETPARAMS, buf, 6);

    // WORKAROUND - Optimizing the Inverted IQ Operation, see datasheet DS_SX1261-2_V1.2
    {
        uint8_t reg_value = 0;

        __sx126x_read_registers(p_this, REG_IQ_POLARITY, &reg_value, 1);

        if (p_this->p_devinfo->p_lora_param->invert_iq == 1) {
            reg_value &= ~( 1 << 2 );  // Bit 2 set to 0 when using inverted IQ polarity
        } else {
            reg_value |= ( 1 << 2 );  // Bit 2 set to 1 when using standard IQ polarity
        }

        __sx126x_write_registers(p_this, REG_IQ_POLARITY, &reg_value, 1);
    }
}

static void __sx126x_get_rx_buffer_status (radio_sx126x_lora_dev_t *p_this, 
                                           uint8_t                 *p_payload_length, 
                                           uint8_t                 *p_rx_start_buf_ptr)
{
    uint8_t status[2] = {0};
    uint8_t reg_val   = 0;

    __sx126x_read_command(p_this, RADIO_GET_RXBUFFERSTATUS, status, 2);

    reg_val = __sx126x_read_register(p_this, REG_LR_PACKETPARAMS) >> 7;
    if (reg_val == 1) {
        *p_payload_length = __sx126x_read_register(p_this, REG_LR_PAYLOADLENGTH);
    } else {
        *p_payload_length = status[0];
    }
    *p_rx_start_buf_ptr = status[1];
}

static int __sx126x_add_registers_to_retention_list(
            radio_sx126x_lora_dev_t *p_this,
            const uint16_t          *register_addr,
            uint8_t                  register_nb)
{
    uint8_t buffer[9] = {0};

    __sx126x_read_registers(p_this, REG_RETENTION_LIST_BASE_ADDRESS, buffer, 9);

    if(1) {
        const uint8_t initial_nb_of_registers = buffer[0];
        uint8_t*      register_list           = &buffer[1];

        for( uint8_t index = 0; index < register_nb; index++)
        {
            bool register_has_to_be_added = true;

            // Check if the current register is already added to the list
            for( uint8_t i = 0; i < buffer[0]; i++ )
            {
                if( register_addr[index] == ( ( uint16_t ) register_list[2 * i] << 8 ) + register_list[2 * i + 1])
                {
                    register_has_to_be_added = false;
                    break;
                }
            }

            if( register_has_to_be_added == true )
            {
                if( buffer[0] < SX126X_MAX_NB_REG_IN_RETENTION )
                {
                    register_list[2 * buffer[0]]     = ( uint8_t )( register_addr[index] >> 8 );
                    register_list[2 * buffer[0] + 1] = ( uint8_t )( register_addr[index] >> 0 );
                    buffer[0] += 1;
                }
                else
                {
                    return  -1;
                }
            }
        }

        if( buffer[0] != initial_nb_of_registers )
        {
            __sx126x_write_registers(p_this, REG_RETENTION_LIST_BASE_ADDRESS, buffer, 9 );
        }
    }

    return  0;
}

static int __sx126x_init_retention_list(radio_sx126x_lora_dev_t *p_this)
{
    const uint16_t list_of_registers[] = {REG_RX_GAIN, REG_TX_MODULATION, REG_IQ_POLARITY};

    return __sx126x_add_registers_to_retention_list(p_this,
                                                    list_of_registers,
                                                    sizeof( list_of_registers ) / sizeof( list_of_registers[0] ) );
}

static int __sx126x_set_sleep (void *p_drv)
{
    radio_sx126x_lora_dev_t *p_this     = (radio_sx126x_lora_dev_t *)p_drv;
    uint8_t                  value      = 0;
    uint8_t                  wakeup_rtc = 0;
    uint8_t                  reset      = 0;
    uint8_t                  warm_start = 1;
    uint8_t                  reserved   = 0;

    __sx126x_init_retention_list(p_this);

    value = (reserved << 3) | (warm_start << 2) | (reset << 1) | (wakeup_rtc);
    __sx126x_write_command(p_this, RADIO_SET_SLEEP, &value, 1);
    __sx126x_delay_ms(p_this, 1);
    SM_STATE_SET(p_this, RFLR_STATE_SLEEP);

    return RADIO_RET_OK;
}



static int __sx126x_set_rx_mode (radio_sx126x_lora_dev_t *p_this)
{
    uint8_t  buf[3]     = {0};
    uint32_t rx_timeout = 0;

    if (p_this->p_gpio->pfn_txen_pin_set) {
        p_this->p_gpio->pfn_txen_pin_set(0);
    }
    if (p_this->p_gpio->pfn_rxen_pin_set) {
        p_this->p_gpio->pfn_rxen_pin_set(1);
    }
    
    __sx126x_set_cap_val_of_xtal(p_this);

    __sx126x_set_stop_rxtimer_on_preamble_detect(p_this, false);
    __sx126x_set_lora_symb_num_timeout(p_this, 0);

    SM_STATE_SET(p_this, RFLR_STATE_RX_RUNNING);

    __sx126x_lora_set_packet_params(p_this);
    __sx126x_set_dio_irq_params(p_this, 
                                IRQ_RX_DONE | IRQ_RX_TX_TIMEOUT | IRQ_CRC_ERROR, 
                                IRQ_RX_DONE | IRQ_RX_TX_TIMEOUT | IRQ_CRC_ERROR, 
                                IRQ_RADIO_NONE,
                                IRQ_RADIO_NONE);

    __sx126x_write_register(p_this, REG_RX_GAIN, 0x96);

    rx_timeout = p_this->p_devinfo->p_lora_param->rx_pkt_timeout * 64; 

    if (p_this->p_devinfo->p_lora_param->rx_mode == 0) {  
        rx_timeout = 0xFFFFFF;   
    } 

    buf[0] = (uint8_t)((rx_timeout >> 16) & 0xFF);
    buf[1] = (uint8_t)((rx_timeout >> 8) & 0xFF);
    buf[2] = (uint8_t)(rx_timeout & 0xFF);

    __sx126x_write_command(p_this, RADIO_SET_RX, buf, 3); 

    return RADIO_RET_OK;
}







static int __sx126x_set_tx_continuous_wave (void *p_drv)
{
    radio_sx126x_lora_dev_t *p_this = (radio_sx126x_lora_dev_t *)p_drv;

    if (p_this->p_gpio->pfn_txen_pin_set) {
        p_this->p_gpio->pfn_txen_pin_set(1);
    }
    if (p_this->p_gpio->pfn_rxen_pin_set) {
        p_this->p_gpio->pfn_rxen_pin_set(0);
    }
    __sx126x_set_cap_val_of_xtal(p_this);

    __sx126x_freq_set(p_this, p_this->p_devinfo->p_lora_param->frequency);
    __sx126x_set_rf_power(p_this, p_this->p_devinfo->p_lora_param->power, RADIO_RAMP_800_US); 

    __sx126x_write_command(p_this, RADIO_SET_TXCONTINUOUSWAVE, 0, 0);
    SM_STATE_SET(p_this, RFLR_STATE_TX_RUNNING);
    return RADIO_RET_OK;
}


static int __sx126x_get_irq_status (void *p_drv)
{
    radio_sx126x_lora_dev_t *p_this        = (radio_sx126x_lora_dev_t *)p_drv;
    uint8_t                  irq_status[2] = {0};

    __sx126x_read_command(p_this, RADIO_GET_IRQSTATUS, irq_status, 2);

    return (irq_status[0] << 8) | irq_status[1];
}


static void __sx126x_clear_irq_status (void *p_drv, uint16_t irq)
{
    radio_sx126x_lora_dev_t *p_this = (radio_sx126x_lora_dev_t *)p_drv;
    uint8_t                  buf[2] = {0};

    buf[0] = (uint8_t)(((uint16_t)irq >> 8) & 0x00FF);
    buf[1] = (uint8_t)((uint16_t)irq & 0x00FF);
    __sx126x_write_command(p_this, RADIO_CLR_IRQSTATUS, buf, 2);
}






static int __sx126x_reset (void *p_arg)
{
    radio_sx126x_lora_dev_t *p_this = (radio_sx126x_lora_dev_t *)p_arg;
    if (p_this->p_gpio->pfn_reset_pin_set == NULL) {
        return -1;
    }

    __sx126x_delay_ms(p_this, 10);
    p_this->p_gpio->pfn_reset_pin_set(0);
    __sx126x_delay_ms(p_this, 20);
    p_this->p_gpio->pfn_reset_pin_set(1);
    __sx126x_delay_ms(p_this, 10);

    return RADIO_RET_OK;
}



static int __sx126x_sync_id_set (void *p_arg, uint8_t *p_id, uint8_t len)
{
    //radio_sx126x_lora_dev_t *p_this = (radio_sx126x_lora_dev_t *)p_arg;
    return RADIO_RET_MODULE_NOT_SUPPORT;
}
    


static int __sx126x_sync_id_get (void *p_arg, uint8_t *p_id, uint8_t *p_len)
{
    //radio_sx126x_lora_dev_t *p_this = (radio_sx126x_lora_dev_t *)p_arg;
    return RADIO_RET_MODULE_NOT_SUPPORT;
}



static int __sx126x_preamble_length_set (void *p_arg, uint16_t preamble_len)
{
    radio_sx126x_lora_dev_t *p_this = (radio_sx126x_lora_dev_t *)p_arg;
    uint8_t                  sf     = p_this->p_devinfo->p_lora_param->sf;
    
    p_this->p_devinfo->p_lora_param->preamble_len = preamble_len;
    if ((sf == 5) || (sf == 6)) {
        if(preamble_len < 12) {
            p_this->p_devinfo->p_lora_param->preamble_len = 12;
        }
    } 
    __sx126x_lora_set_packet_params(p_this);
    
    return RADIO_RET_OK;
}



static int __sx126x_preamble_length_get (void *p_arg, uint16_t *p_len)
{
    radio_sx126x_lora_dev_t *p_this = (radio_sx126x_lora_dev_t *)p_arg;
    uint8_t                  sf     = p_this->p_devinfo->p_lora_param->sf;
    
    *p_len = p_this->p_devinfo->p_lora_param->preamble_len;
    if ((sf == 5) || (sf == 6)) {
        if(*p_len < 12) {
            *p_len = 12;
        }
    } 
    return RADIO_RET_OK;
}



static int __sx126x_set_rf_power (radio_sx126x_lora_dev_t *p_this, 
                                  int8_t                   power, 
                                  uint8_t                  ramp_time)
{
    uint8_t  buf[2] = {0};
    uint8_t  pa_dc  = 0x04;   /* PAռձ Ĭ0x04 */
    uint8_t  hp_mx  = 0x07;   /* PAĴС Ĭ0x07 */
    
    __sx126x_set_pa_config(p_this, pa_dc, hp_mx, 0x00, 0x01); 

    if ((power < -9) || ((power > 22))) {
        return RADIO_RET_PARAM_INVALID;
    }
    
    __sx126x_write_register(p_this, REG_OCP_CONFIG, 0x38);

    buf[0] = power;
    buf[1] = ramp_time;

    p_this->p_devinfo->p_lora_param->power = power;
    __sx126x_write_command(p_this, RADIO_SET_TXPARAMS, buf, 2);
    
    return RADIO_RET_OK;
}

static int __sx126x_power_set_adapter (void *p_arg, uint8_t level, float *p_power)
{
    uint8_t                  ret;
    radio_sx126x_lora_dev_t *p_this = (radio_sx126x_lora_dev_t *)p_arg;
    float power_tab[] = {-9, -5, -1, 3, 7, 12, 17, 22};  /* 8ʵȼ0 ~ 7 */

    if (level > (sizeof(power_tab) / sizeof(power_tab[0]))) {
        return RADIO_RET_CFG_OUT_OF_RANGE;
    }
 
    ret = __sx126x_set_rf_power(p_this, power_tab[level], RADIO_RAMP_800_US);
    if (ret == RADIO_RET_OK) {
        *p_power = p_this->p_devinfo->p_lora_param->power = power_tab[level];
    } else {
        return RADIO_RET_PARAM_INVALID;
    }
    
    return RADIO_RET_OK;
}



static int __sx126x_freq_set (void *p_arg, uint32_t frequency)
{
    radio_sx126x_lora_dev_t *p_this         = (radio_sx126x_lora_dev_t *)p_arg;
    uint8_t                 buf[4]          = {0};
    uint32_t                freq            = 0;
    static bool             ImageCalibrated = false;
    
    if ((frequency < p_this->p_devinfo->p_lora_param->dev_type.min_freq) || 
        (frequency > p_this->p_devinfo->p_lora_param->dev_type.max_freq)) {
        return RADIO_RET_PARAM_INVALID;
    }

    if (ImageCalibrated == false) {
        __sx126x_calibrate_image(p_this, frequency);
        ImageCalibrated = true;
    }

    freq = (uint32_t)((double)frequency / (double)FREQ_STEP);
    buf[0] = (uint8_t)((freq >> 24) & 0xFF);
    buf[1] = (uint8_t)((freq >> 16) & 0xFF);
    buf[2] = (uint8_t)((freq >> 8) & 0xFF);
    buf[3] = (uint8_t)(freq & 0xFF);

    p_this->p_devinfo->p_lora_param->frequency = frequency;
    __sx126x_write_command (p_this, RADIO_SET_RFFREQUENCY, buf, 4);
    return RADIO_RET_OK;
}



static int __sx126x_mode_set (void *p_arg, uint8_t mode)
{
    radio_sx126x_lora_dev_t *p_this = (radio_sx126x_lora_dev_t *)p_arg;
    int                      ret    = RADIO_RET_OK;

    switch (mode) {

    case SLEEP_MODE:
        ret = __sx126x_set_sleep(p_this);
        break;

    case STANDBY_MODE:
        ret = __sx126x_set_standby(p_this, STDBY_XOSC_MODE);
        break;

    case RX_MODE:
        ret = __sx126x_set_rx_mode(p_this);
        break;

    case CW_MODE:  
        ret = __sx126x_set_tx_continuous_wave(p_this);
        break;

    case RX_BOOST_MODE:
        ret = __sx126x_radio_rx_boosted(p_this, 20000);
        break;

    case RX_DC_MODE:
        ret = __sx126x_set_rx_duty_cycle(p_this, 10000, 10000);
        break;

    case FS_MODE:
        ret = __sx126x_set_fs(p_this);
        break;

    default:
        break;
    }

    return ret;
}

 
static int __sx126x_state_get (void *p_arg, radio_state_t *p_state)
{
    radio_sx126x_lora_dev_t *p_this = (radio_sx126x_lora_dev_t *)p_arg;

    switch (SM_STATE_GET(p_this)) {
    case RFLR_STATE_TX_RUNNING:
    case RFLR_STATE_TX_INIT:
        *p_state = TX_RUNING_ST;
        break;

    case RFLR_STATE_TX_DONE:
    case RFLR_STATE_TX_TIMEOUT:
        *p_state = IDLE_ST;
        break;

    case RFLR_STATE_RX_DONE:
    case RFLR_STATE_RX_TIMEOUT:
        if (p_this->p_devinfo->p_lora_param->rx_mode == CONTINUOUS_RECV_MODE) {
            *p_state = RX_RUNING_ST;
        } else {
            *p_state = IDLE_ST;
        }
        break;

    case RFLR_STATE_RX_INIT:
    case RFLR_STATE_RX_RUNNING:
        *p_state = RX_RUNING_ST;
        break;

    case RFLR_STATE_CAD_RUNNING:
    case RFLR_STATE_CAD_INIT:
        *p_state = CAD_ST;
        break;

    case RFLR_STATE_IDLE:   
        *p_state = IDLE_ST;  
        break;

    default:
        *p_state = IDLE_ST;
        break;
    }

    return RADIO_RET_OK;
}



static int __sx126x_cur_rssi_get (void *p_arg, int16_t *p_rssi)
{
    radio_sx126x_lora_dev_t *p_this = (radio_sx126x_lora_dev_t *)p_arg;
    uint8_t                  buf[1] = {0};

    __sx126x_read_command(p_this, RADIO_GET_RSSIINST, buf, 1);
    *p_rssi = (int16_t)(-buf[0] >> 1);

    return RADIO_RET_OK;
}






static int __sx126x_time_on_air_get (void     *p_arg, 
                                     uint16_t  preamble_len,
                                     uint8_t   pkt_len,
                                     uint32_t *p_time)
{
    radio_sx126x_lora_dev_t *p_this = (radio_sx126x_lora_dev_t *)p_arg;
    uint8_t                  ldro   = 0;
    uint8_t                  sf     = p_this->p_devinfo->p_lora_param->sf;
    uint8_t                  bw     = p_this->p_devinfo->p_lora_param->bandwidth;

    if (preamble_len == RADIO_TIME_ON_AIR_USE_CUR_PREAMBLE_LEN) {
        preamble_len = p_this->p_devinfo->p_lora_param->preamble_len;
        if ((sf == 5) || (sf == 6)) {
            if (preamble_len < 12) {
                preamble_len = 12;
            }
        } 
    }
    
                                  /*  SF12    SF11    SF10   SF9    SF8    SF7    SF6    SF5*/
    double lora_symb_time[4][8] = {{ 65.536, 32.768, 16.384, 8.192, 4.096, 2.048, 1.024, 0.512 },  /* 6 -- 62.5 KHz */
                                   { 32.768, 16.384, 8.192,  4.096, 2.048, 1.024, 0.512, 0.256 },  /* 7 -- 125  KHz */
                                   { 16.384, 8.192,  4.096,  2.048, 1.024, 0.512, 0.256, 0.128 },  /* 8 -- 250  KHz */
                                   { 8.192,  4.096,  2.048,  1.024, 0.512, 0.256, 0.128, 0.064 }}; /* 9 -- 500  KHz */

    double ts = lora_symb_time[bw - 6][12 - sf];

    double t_preamble = (preamble_len + 4.25) * ts;   

    if (p_this->p_devinfo->p_lora_param->ldro_auto == 1) {
        if (((bw == 6) && ((sf == 10) || (sf == 11)   || (sf == 12))) ||
            ((bw == 7) && ((sf == 11) || (sf == 12))) ||
            ((bw == 8) &&  (sf == 12))) {
            ldro = 0x01;
        } else {
            ldro = 0x00;
        }
    } else {
        ldro = !!p_this->p_devinfo->p_lora_param->ldro_enable;
    }

    double tmp = ceil((8 * pkt_len - 4 * p_this->p_devinfo->p_lora_param->sf +
                      28 + 16 * p_this->p_devinfo->p_lora_param->crc_on -
                      (( p_this->p_devinfo->p_lora_param->header_mode == LORA_PACKET_FIXED_LENGTH) ? 20 : 0)) /
                      (double)(4 * ( p_this->p_devinfo->p_lora_param->sf -
                      ((ldro > 0) ? 2 : 0)))) *
                      ((p_this->p_devinfo->p_lora_param->coderate % 4) + 4);     

    double n_payload = 8 + ((tmp > 0) ? tmp : 0);
    double t_payload = n_payload * ts;
    double t_on_air = t_preamble + t_payload;
                      
    *p_time = floor(t_on_air + 0.999);

    return RADIO_RET_OK;
}






static int __sx126x_data_send (void *p_arg, uint8_t *p_buf, uint8_t size)
{
    radio_sx126x_lora_dev_t *p_this       = (radio_sx126x_lora_dev_t *)p_arg;
    uint8_t                  buf[3]       = {0};
    uint16_t                 irq_regs     = 0;
    uint32_t                 timeout      = 0;
    uint32_t                 time_on_air  = 0;
    uint16_t                 crc          = 0;
    uint8_t                  crc_bytes[2] = {0};


    if ((NULL == p_buf)) {
        return RADIO_RET_PARAM_INVALID;
    }

    if (p_this->p_gpio->pfn_txen_pin_set) {
        p_this->p_gpio->pfn_txen_pin_set(1);
    }
    if (p_this->p_gpio->pfn_rxen_pin_set) {
        p_this->p_gpio->pfn_rxen_pin_set(0);
    }
    
    if (p_this->p_devinfo->p_lora_param->crc_software == 1) {
        if (size > 253) {
            return RADIO_RET_CFG_OUT_OF_RANGE;
        }
        crc = __crc16_ccitt(p_buf, size);
        crc_bytes[0] = (crc & 0xFF00) >> 8;
        crc_bytes[1] = crc & 0xFF;
        size += 2;
    } else {
        if (size > 255) {
            return RADIO_RET_CFG_OUT_OF_RANGE;
        }
    }

    __sx126x_set_cap_val_of_xtal(p_this);

    __sx126x_set_dio_irq_params(p_this, 
                                IRQ_TX_DONE | IRQ_RX_TX_TIMEOUT,
                                IRQ_RADIO_NONE,
                                IRQ_RADIO_NONE,
                                IRQ_RADIO_NONE);
    
    if (p_this->p_devinfo->p_lora_param->header_mode == 0) {
        /* Ϊʽͷģʽͨsize޸ȫֵpayload_len֪LoRaҪ͵ֽ*/
        p_this->p_devinfo->p_lora_param->payload_len = size;
    } else {
        /* Ϊʽͷֽpayload_lenΪsx126x_radio_lora_cfg.c³ʼʱõĹ̶ֵ */
        /* ܴڡsize != payload_len⣬Ҫûд */
    }

    SM_STATE_SET(p_this, RFLR_STATE_TX_RUNNING);
    __sx126x_lora_set_packet_params(p_this);
    if (p_this->p_devinfo->p_lora_param->crc_software == 1) {
        __sx126x_write_buffer(p_this, 0x00, p_buf, size - 2);
        __sx126x_write_buffer(p_this, 0x00 + size - 2, crc_bytes, 2);
    } else {
        __sx126x_write_buffer(p_this, 0x00, p_buf, size);
    }
    
    timeout = p_this->p_devinfo->p_lora_param->tx_timeout;
    __sx126x_time_on_air_get(p_this, 
                             RADIO_TIME_ON_AIR_USE_CUR_PREAMBLE_LEN, 
                             size, 
                             &time_on_air);

    if (timeout == 1) {
        timeout = (time_on_air + 20) * 64;
    } else {
        timeout = 0;
    }

    buf[0] = (uint8_t)((timeout >> 16) & 0xFF);
    buf[1] = (uint8_t)((timeout >> 8) & 0xFF);
    buf[2] = (uint8_t)(timeout & 0xFF);
    __sx126x_write_command(p_this, RADIO_SET_TX, buf, 3);
  
    while (1) {
        
        irq_regs = __sx126x_get_irq_status(p_this);
        
        if ((irq_regs & IRQ_TX_DONE) == IRQ_TX_DONE) { 
            __sx126x_clear_irq_status(p_this, IRQ_RADIO_ALL);
            SM_STATE_SET(p_this, RFLR_STATE_TX_DONE);
            if (p_this->p_devinfo->p_events->pfn_tx_done_cb != NULL) {
                p_this->p_devinfo->p_events->pfn_tx_done_cb();
            }
            break;
        }
        
        if ((irq_regs & IRQ_RX_TX_TIMEOUT ) == IRQ_RX_TX_TIMEOUT) {
            __sx126x_clear_irq_status(p_this, IRQ_RADIO_ALL);
            SM_STATE_SET(p_this, RFLR_STATE_TX_TIMEOUT);  
            if (p_this->p_devinfo->p_events->pfn_tx_timeout_cb != NULL) {
                p_this->p_devinfo->p_events->pfn_tx_timeout_cb();
            }
            return RADIO_RET_TX_TIMEOUT;
        }
        
    }

    return RADIO_RET_OK;
}





static int __sx126x_rx_state_check (void *p_arg, int *p_rx_state)
{
    radio_sx126x_lora_dev_t *p_this   = (radio_sx126x_lora_dev_t *)p_arg;
    uint16_t                 irq_regs = 0;

    irq_regs = __sx126x_get_irq_status(p_this);

    if ((irq_regs & IRQ_RX_TX_TIMEOUT) == IRQ_RX_TX_TIMEOUT) {    /* ճʱ */
        if (p_this->p_devinfo->p_events->pfn_rx_timeout_cb != NULL) {
            p_this->p_devinfo->p_events->pfn_rx_timeout_cb();   
        }
        *p_rx_state = RADIO_RET_RX_TIMEOUT;
    } else if ((irq_regs & IRQ_CRC_ERROR) == IRQ_CRC_ERROR) {    /* ӲCRC */
        if (p_this->p_devinfo->p_events->pfn_rx_hw_crc_error_cb != NULL) {
            p_this->p_devinfo->p_events->pfn_rx_hw_crc_error_cb();   
        } 
        *p_rx_state = RADIO_RET_RX_CRC_ERR;
    } else if ((irq_regs & IRQ_RX_DONE) == IRQ_RX_DONE) {    /*  */
        *p_rx_state = RADIO_RET_OK;
    } else {
        *p_rx_state = RADIO_RET_NOT_RECV_PKT;
    }
    
    /* ж֮Ŷжϱ־ */
    if (*p_rx_state != RADIO_RET_NOT_RECV_PKT) {
        __sx126x_clear_irq_status(p_this, IRQ_RADIO_ALL); 
    }
    
    return RADIO_RET_OK;
}





static int __sx126x_data_recv (void *p_arg, uint8_t *p_buf, uint8_t *p_size)
{
    radio_sx126x_lora_dev_t *p_this    = (radio_sx126x_lora_dev_t *)p_arg;
    int                      rx_state  = 0;
    uint8_t                  offset    = 0;
    uint8_t                  status[3] = {0};
    int16_t                  pkt_rssi  = 0;
    int16_t                  pkt_snr   = 0;
    uint8_t                  pkt_size  = 0; 
    uint16_t                 crc       = 0;

    if (SM_STATE_GET(p_this) != RFLR_STATE_RX_RUNNING) {
        return RADIO_RET_NOT_RECV_PKT;
    }
    
    if ((p_buf == NULL) || (p_size == NULL)) {
        return RADIO_RET_PARAM_INVALID;
    }

    __sx126x_rx_state_check(p_this, &rx_state);

    if ((rx_state == RADIO_RET_RX_TIMEOUT) || (rx_state == RADIO_RET_RX_CRC_ERR)) {
        *p_size = 0;
        if (p_this->p_devinfo->p_lora_param->rx_mode == CONTINUOUS_RECV_MODE) {
            SM_STATE_SET(p_this, RFLR_STATE_RX_RUNNING);
        } else {
            SM_STATE_SET(p_this, RFLR_STATE_IDLE);
        }
        return rx_state;
    } else if (rx_state == RADIO_RET_NOT_RECV_PKT) {
        SM_STATE_SET(p_this, RFLR_STATE_RX_RUNNING);   
        return rx_state;
    } else {
        SM_STATE_SET(p_this, RFLR_STATE_RX_DONE);
    }
    
    __sx126x_get_rx_buffer_status(p_this, &pkt_size, &offset);
    
    if (p_this->p_devinfo->p_lora_param->crc_software) {   
        if (pkt_size < 3) {
            goto sw_crc_err;
        } 
    } else {
        if (pkt_size < 1) {
            goto sw_crc_err;
        }
    }
    
    __sx126x_read_buffer(p_this, offset, p_buf, pkt_size);
    *p_size = pkt_size;

    if (p_this->p_devinfo->p_lora_param->crc_software == 1) {
        crc = __crc16_ccitt(p_buf, *p_size - 2);
        if ((((crc & 0xFF00) >> 8) != p_buf[*p_size - 2]) || 
             ((crc & 0xFF) != p_buf[*p_size - 1])) {
            goto sw_crc_err;
        }
        *p_size -= 2;
    }

    __sx126x_read_command(p_this, RADIO_GET_PACKETSTATUS, status, 3);

    pkt_rssi = -status[0] >> 1;
    (status[1] < 128) ? (pkt_snr = status[1] >> 2) : (pkt_snr = ((status[1] - 256 ) >> 2));

    if (p_this->p_devinfo->p_events != NULL) {
        if (p_this->p_devinfo->p_events->pfn_rx_done_cb != NULL) {
            p_this->p_devinfo->p_events->pfn_rx_done_cb(p_buf,
                                                        *p_size, 
                                                        pkt_rssi, 
                                                        pkt_snr);
        }
    }

    if (p_this->p_devinfo->p_lora_param->rx_mode == CONTINUOUS_RECV_MODE) {
        SM_STATE_SET(p_this, RFLR_STATE_RX_RUNNING);
    } else {
        SM_STATE_SET(p_this, RFLR_STATE_IDLE);
    }

    return RADIO_RET_OK;

sw_crc_err:
    *p_size = 0;
    if (p_this->p_devinfo->p_events->pfn_rx_sw_crc_error_cb != NULL) {
        p_this->p_devinfo->p_events->pfn_rx_sw_crc_error_cb();
    }
    if (p_this->p_devinfo->p_lora_param->rx_mode == CONTINUOUS_RECV_MODE) {
        SM_STATE_SET(p_this, RFLR_STATE_RX_RUNNING);
    } else {
        SM_STATE_SET(p_this, RFLR_STATE_IDLE);
    }
    return RADIO_RET_RX_CRC_ERR;
}

static int __sx126x_ioctl (void *p_arg, int cmd, void *p_param)
{
    radio_sx126x_lora_dev_t *p_this = (radio_sx126x_lora_dev_t *)p_arg;
    int                      ret    = RADIO_RET_OK;
    uint8_t                  sf     = p_this->p_devinfo->p_lora_param->sf;
    uint8_t                  bw     = p_this->p_devinfo->p_lora_param->bandwidth;
    uint8_t                  cr     = p_this->p_devinfo->p_lora_param->coderate;

    switch (cmd) {

    case SET_CAD_ONLY:
        __sx126x_set_cad(p_this, LORA_CAD_ONLY, 0);
        break;

    case SET_CAD_RX:
        __sx126x_set_cad(p_this, LORA_CAD_RX, *(uint32_t *)p_param);
        break;

    case CAD_START:
        ret = __sx126x_cad_start(p_this, (uint8_t *)p_param);
        break;

    case BW_SET:
        bw = *(uint8_t *)p_param;
        if (__sx126x_lora_set_modulation_params(p_this, sf, bw, cr) == RADIO_RET_OK) {
            p_this->p_devinfo->p_lora_param->bandwidth = bw;
        }
        break;

    case SF_SET:
        sf = *(uint8_t *)p_param;
        if (__sx126x_lora_set_modulation_params(p_this, sf, bw, cr) == RADIO_RET_OK) {
            p_this->p_devinfo->p_lora_param->sf = sf; 
        }
        break;
        
    case CAP_SET:
        p_this->p_devinfo->p_lora_param->dev_type.cap_val = *(uint8_t *)p_param;
        break;
        
    default:
        break;
    }

    return ret;
}


static void __sx126x_dio1_irq_func (void *p_arg)
{ 
}


static int __sx126x_cad_param_match (radio_sx126x_lora_dev_t *p_this,
                                     uint8_t                 *p_cad_symbol_num,
                                     uint8_t                 *p_cad_det_peak,
                                     uint8_t                 *p_cad_det_min)
{
    uint8_t sf = p_this->p_devinfo->p_lora_param->sf;
    uint8_t bw = p_this->p_devinfo->p_lora_param->bandwidth;

    *p_cad_det_min = 10;

    switch (bw) {
    case 6:
        switch (sf) {
        case 5:  *p_cad_det_peak = 24; *p_cad_symbol_num = LORA_CAD_02_SYMBOL; break;
        case 6:  *p_cad_det_peak = 25; *p_cad_symbol_num = LORA_CAD_02_SYMBOL; break;
        case 7:  *p_cad_det_peak = 25; *p_cad_symbol_num = LORA_CAD_02_SYMBOL; break;
        case 8:  *p_cad_det_peak = 25; *p_cad_symbol_num = LORA_CAD_04_SYMBOL; break;
        case 9:  *p_cad_det_peak = 26; *p_cad_symbol_num = LORA_CAD_04_SYMBOL; break;
        case 10: *p_cad_det_peak = 26; *p_cad_symbol_num = LORA_CAD_04_SYMBOL; break;
        case 11: *p_cad_det_peak = 27; *p_cad_symbol_num = LORA_CAD_04_SYMBOL; break;
        case 12: *p_cad_det_peak = 29; *p_cad_symbol_num = LORA_CAD_04_SYMBOL; break;
        default: break;
        }
        break;

    case 7:
        switch (sf) {
        case 5:  *p_cad_det_peak = 24; *p_cad_symbol_num = LORA_CAD_02_SYMBOL; break;
        case 6:  *p_cad_det_peak = 24; *p_cad_symbol_num = LORA_CAD_04_SYMBOL; break;
        case 7:  *p_cad_det_peak = 25; *p_cad_symbol_num = LORA_CAD_04_SYMBOL; break;
        case 8:  *p_cad_det_peak = 25; *p_cad_symbol_num = LORA_CAD_04_SYMBOL; break;
        case 9:  *p_cad_det_peak = 26; *p_cad_symbol_num = LORA_CAD_04_SYMBOL; break;
        case 10: *p_cad_det_peak = 27; *p_cad_symbol_num = LORA_CAD_04_SYMBOL; break;
        case 11: *p_cad_det_peak = 27; *p_cad_symbol_num = LORA_CAD_04_SYMBOL; break;
        case 12: *p_cad_det_peak = 29; *p_cad_symbol_num = LORA_CAD_04_SYMBOL; break;
        default:
            break;
        }
        break;

    case 8:
        switch (sf) {
        case 5:  *p_cad_det_peak = 24; *p_cad_symbol_num = LORA_CAD_02_SYMBOL; break;
        case 6:  *p_cad_det_peak = 25; *p_cad_symbol_num = LORA_CAD_02_SYMBOL; break;
        case 7:  *p_cad_det_peak = 25; *p_cad_symbol_num = LORA_CAD_02_SYMBOL; break;
        case 8:  *p_cad_det_peak = 26; *p_cad_symbol_num = LORA_CAD_04_SYMBOL; break;
        case 9:  *p_cad_det_peak = 27; *p_cad_symbol_num = LORA_CAD_04_SYMBOL; break;
        case 10: *p_cad_det_peak = 27; *p_cad_symbol_num = LORA_CAD_04_SYMBOL; break;
        case 11: *p_cad_det_peak = 27; *p_cad_symbol_num = LORA_CAD_04_SYMBOL; break;
        case 12: *p_cad_det_peak = 30; *p_cad_symbol_num = LORA_CAD_04_SYMBOL; break;
        default: break;
        }
        break;

    case 9:
        switch (sf) {
        case 5:  *p_cad_det_peak = 24; *p_cad_symbol_num = LORA_CAD_04_SYMBOL; break;
        case 6:  *p_cad_det_peak = 24; *p_cad_symbol_num = LORA_CAD_04_SYMBOL; break;
        case 7:  *p_cad_det_peak = 25; *p_cad_symbol_num = LORA_CAD_04_SYMBOL; break;
        case 8:  *p_cad_det_peak = 27; *p_cad_symbol_num = LORA_CAD_04_SYMBOL; break;
        case 9:  *p_cad_det_peak = 30; *p_cad_symbol_num = LORA_CAD_04_SYMBOL; break;
        case 10: *p_cad_det_peak = 30; *p_cad_symbol_num = LORA_CAD_04_SYMBOL; break;
        case 11: *p_cad_det_peak = 30; *p_cad_symbol_num = LORA_CAD_08_SYMBOL; break;
        case 12: *p_cad_det_peak = 31; *p_cad_symbol_num = LORA_CAD_08_SYMBOL; break;
        default: break;
        }
        break;

    default:
        break;
    }

    return RADIO_RET_OK;
}



static int __sx126x_set_cad (radio_sx126x_lora_dev_t *p_this, uint8_t cad_mode, uint32_t rx_timeout)
{
    uint8_t  buf[7]         = {0};
    uint8_t  cad_symbol_num = 0;
    uint8_t  cad_det_peak   = 0;
    uint8_t  cad_det_min    = 0;

    uint32_t cad_rx_timeout = 0;
    
    cad_rx_timeout = rx_timeout << 6;
    
    __sx126x_set_cap_val_of_xtal(p_this);

    if (cad_mode == LORA_CAD_ONLY) {
        __sx126x_set_dio_irq_params(p_this,
                                    IRQ_CAD_DONE | IRQ_CAD_ACTIVITY_DETECTED,
                                    IRQ_RADIO_NONE,    
                                    IRQ_RADIO_NONE,
                                    IRQ_RADIO_NONE);
    } else if (cad_mode == LORA_CAD_RX) {
        __sx126x_set_dio_irq_params(p_this,
                                    IRQ_CAD_DONE              | 
                                    IRQ_CAD_ACTIVITY_DETECTED | 
                                    IRQ_RX_DONE               | 
                                    IRQ_RX_TX_TIMEOUT         | 
                                    IRQ_CRC_ERROR,
                                    IRQ_RX_DONE       | 
                                    IRQ_RX_TX_TIMEOUT | 
                                    IRQ_CRC_ERROR,
                                    IRQ_RADIO_NONE,
                                    IRQ_RADIO_NONE);
    }

    __sx126x_cad_param_match(p_this, &cad_symbol_num, &cad_det_peak, &cad_det_min);
    
    buf[0] = cad_symbol_num;
    buf[1] = cad_det_peak;
    buf[2] = cad_det_min;
    buf[3] = cad_mode;     
    buf[4] = (uint8_t)((cad_rx_timeout >> 16) & 0xFF);
    buf[5] = (uint8_t)((cad_rx_timeout >> 8) & 0xFF);
    buf[6] = (uint8_t)(cad_rx_timeout & 0xFF);

    __sx126x_write_command(p_this, RADIO_SET_CADPARAMS, buf, 7);

    return RADIO_RET_OK;
}



static int __sx126x_cad_start (radio_sx126x_lora_dev_t *p_this, uint8_t *p_state)
{
    uint16_t irq_regs = 0;
    
    SM_STATE_SET(p_this, RFLR_STATE_CAD_RUNNING);
    CAD_STATE_SET(p_this, CAD_STATE_START);
    __sx126x_write_command(p_this, RADIO_SET_CAD, 0, 0);  

    while(1) {

        irq_regs = __sx126x_get_irq_status(p_this);
        
        if ((irq_regs & IRQ_CAD_DONE) == IRQ_CAD_DONE) {  
        
            __sx126x_clear_irq_status(p_this, IRQ_CAD_DONE);
            irq_regs = __sx126x_get_irq_status(p_this);
            CAD_STATE_SET(p_this, CAD_STATE_DONE);

            if (p_this->p_devinfo->p_events->pfn_cad_done_cb != NULL) {
                 p_this->p_devinfo->p_events->pfn_cad_done_cb(true);
            }

            if ((irq_regs & IRQ_CAD_ACTIVITY_DETECTED ) == IRQ_CAD_ACTIVITY_DETECTED) {
                __sx126x_clear_irq_status(p_this, IRQ_CAD_ACTIVITY_DETECTED);
                *p_state = CAD_DETECT;  
                CAD_STATE_SET(p_this, CAD_STATE_DETECT);
                SM_STATE_SET(p_this, RFLR_STATE_RX_RUNNING);  
            } else {
                __sx126x_clear_irq_status(p_this, IRQ_RADIO_ALL);
                *p_state = CAD_NONE;
                CAD_STATE_SET(p_this, CAD_STATE_FINISH);
                 __sx126x_mode_set(p_this, STANDBY_MODE);
                SM_STATE_SET(p_this, RFLR_STATE_IDLE);
            }

            break;

        }

    }

    return RADIO_RET_OK;
}



static const struct radio_drv_funcs __g_sx126x_drv_funcs = 
{
    __sx126x_reset,
    __sx126x_sync_id_set,
    __sx126x_sync_id_get,
    __sx126x_preamble_length_set,
    __sx126x_preamble_length_get,
    __sx126x_power_set_adapter,
    __sx126x_freq_set,
    __sx126x_mode_set,
    __sx126x_state_get,
    __sx126x_cur_rssi_get,
    __sx126x_time_on_air_get,
    __sx126x_data_send,
    __sx126x_data_recv,
    __sx126x_ioctl,
    NULL,
    __sx126x_dio1_irq_func,
    NULL,
    NULL,
    NULL,
};



static int __sx126x_lora_init (radio_sx126x_lora_dev_t *p_this)
{
    int ret;

    SM_STATE_SET(p_this, RFLR_STATE_IDLE);
    
    ret = __sx126x_freq_set(p_this, p_this->p_devinfo->p_lora_param->frequency);
    if (ret != RADIO_RET_OK) {
        return ret;
    }
    
    ret = __sx126x_set_rf_power(p_this, p_this->p_devinfo->p_lora_param->power, RADIO_RAMP_800_US);
    if (ret != RADIO_RET_OK) {
        return ret;
    }
    
    __sx126x_mode_set(p_this, STANDBY_MODE);
    __sx126x_set_lora_mode(p_this);
    
    ret = __sx126x_lora_set_modulation_params(p_this, 
                                              p_this->p_devinfo->p_lora_param->sf, 
                                              p_this->p_devinfo->p_lora_param->bandwidth, 
                                              p_this->p_devinfo->p_lora_param->coderate);
    
    if (ret != RADIO_RET_OK) {
        return ret;
    }
    
    __sx126x_lora_set_packet_params(p_this);

    return RADIO_RET_OK;
}



radio_handle_t radio_sx126x_lora_init (radio_sx126x_lora_dev_t     *p_sx126x,
                                       sx126x_spi_funcs_t          *p_spi,
                                       sx126x_gpio_funcs_t         *p_gpio,
                                       sx126x_delay_funcs_t        *p_delay,
                                       sx126x_cpu_lock_funcs_t     *p_cpu_lock,
                                       radio_sx126x_lora_devinfo_t *p_devinfo)
{
    uint8_t calib_param = 0x7F;
    uint8_t test_ret    = 0;  
    uint8_t val_ant_mis = 0;

    if ((p_sx126x == NULL)||(p_spi == NULL)||(p_devinfo == NULL)) {
        return NULL;
    }
    
    p_sx126x->p_spi        = p_spi;
    p_sx126x->p_gpio       = p_gpio;
    p_sx126x->p_delay      = p_delay;
    p_sx126x->p_lock       = p_cpu_lock;
    p_sx126x->p_devinfo    = p_devinfo;
    p_sx126x->isa.p_funcs  = &__g_sx126x_drv_funcs;
    p_sx126x->isa.p_drv    = p_sx126x;

    SM_STATE_SET(p_sx126x, RFLR_STATE_IDLE);

    __sx126x_reset(p_sx126x);  
    __sx126x_wake_up(p_sx126x);   
    __sx126x_set_standby(p_sx126x, STDBY_RC_MODE);
   
    if (p_sx126x->p_devinfo->p_lora_param->dev_type.osc_type == OSC_TCXO) {
        __sx126x_set_dio3_as_tcxo_ctrl(p_sx126x, TCXO_CTRL_3_3V, 100);
        __sx126x_calibrate(p_sx126x, calib_param);  
    }
    
    __sx126x_write_register(p_sx126x, REG_LR_PAYLOADLENGTH, 0x11);
    test_ret = __sx126x_read_register(p_sx126x, REG_LR_PAYLOADLENGTH);
    if (test_ret != 0x11) {
        return NULL;
    }
    
    __sx126x_set_dio2_as_rf_switch_ctrl(p_sx126x, true);      
    __sx126x_set_standby(p_sx126x, STDBY_RC_MODE);
    __sx126x_set_regulator_mode(p_sx126x, USE_DCDC);  
    __sx126x_set_buffer_base_address(p_sx126x, 0x00, 0x00);    

    if (__sx126x_lora_init(p_sx126x) != RADIO_RET_OK) {
        return NULL;
    }

    val_ant_mis = __sx126x_read_register(p_sx126x, REG_TX_CLAMP_CONFIG);
    val_ant_mis = val_ant_mis | 0x1E;
    __sx126x_write_register(p_sx126x, REG_TX_CLAMP_CONFIG, val_ant_mis);  

    return &p_sx126x->isa;
}



